"""This module contains the default values and constants used in pvfactors """
import numpy as np

# Geometry params
DEFAULT_NORMAL_VEC = None
TOL_COLLINEAR = 1e-5

# Ground params
MAX_X_GROUND = 1e2
MIN_X_GROUND = - MAX_X_GROUND
Y_GROUND = 0.

# PV rows parameters
X_ORIGIN_PVROWS = 0.
DEFAULT_RHO_FRONT = 0.01
DEFAULT_RHO_BACK = 0.03

# Define colors used for plotting the 2D arrays
COLOR_DIC = {
    'i': '#FFBB33',
    's': '#A7A49D',
    't': '#6699cc',
    'pvrow_illum': '#6699cc',
    'pvrow_shaded': '#ff0000',
    'ground_shaded': '#A7A49D',
    'ground_illum': '#FFBB33'
}
PLOT_FONTSIZE = 20
ALPHA_TEXT = 0.20

# Tolerance and thresholds to use from experience getting errors with shapely
DISTANCE_TOLERANCE = 1e-8
THRESHOLD_DISTANCE_TOO_CLOSE = 1e-10


# Gaussian shading default parameters: TOTAL_GAUSSIAN_AREA dependent on these
SIGMA = 1. / np.sqrt(2.)
N_SIGMA = 3.
GAUSSIAN_DIAMETER_CIRCUMSOLAR = 2. * N_SIGMA * SIGMA
RADIUS_CIRCUMSOLAR = 1.
DEFAULT_CIRCUMSOLAR_ANGLE = 30.

# Horizon band shading
DEFAULT_HORIZON_BAND_ANGLE = 6.5

SKY_REFLECTIVITY_DUMMY = 1.



"""Timeseries geometry tools. They allow the vectorization of geometry
calculations."""

import numpy as np
from pvfactors.config import DISTANCE_TOLERANCE
from pvfactors.geometry.base import PVSurface, ShadeCollection
from shapely.geometry import GeometryCollection


class TsShadeCollection(object):
    """Collection of timeseries surfaces that are all either shaded or
    illuminated. This will be used by both ground and PV row
    geometries."""

    def __init__(self, list_ts_surfaces, shaded):
        """Initialize using list of surfaces and shading status

        Parameters
        ----------
        list_ts_surfaces : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsSurface`
            List of timeseries surfaces in collection
        shaded : bool
            Shading status of the collection
        """
        self._list_ts_surfaces = list_ts_surfaces
        self.shaded = shaded
        # TODO: maybe the ts surfaces should have a "shaded" attribute

    @property
    def list_ts_surfaces(self):
        """List of timeseries surfaces in collection"""
        return self._list_ts_surfaces

    @property
    def length(self):
        """Total length of the collection"""
        length = 0.
        for ts_surf in self._list_ts_surfaces:
            length += ts_surf.length
        return length

    @property
    def n_ts_surfaces(self):
        """Number of timeseries surfaces in the collection"""
        return len(self._list_ts_surfaces)

    def get_param_weighted(self, param):
        """Get timeseries parameter for the collection, after weighting by
        surface length.

        Parameters
        ----------
        param : str
            Name of parameter

        Returns
        -------
        np.ndarray
            Weighted parameter values
        """
        return self.get_param_ww(param) / self.length

    def get_param_ww(self, param):
        """Get timeseries parameter from the collection with weight,
        i.e. after multiplying by the surface lengths.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        np.ndarray
            Timeseries parameter values multiplied by weights
        """

        value = 0
        for ts_surf in self._list_ts_surfaces:
            value += ts_surf.length * ts_surf.get_param(param)
        return value

    def update_params(self, new_dict):
        """Update timeseries surface parameters of the segment.

        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surfaces
        """
        for ts_surf in self._list_ts_surfaces:
            ts_surf.params.update(new_dict)

    def at(self, idx):
        """Generate a ponctual shade collection for the desired index.

        Parameters
        ----------
        idx : int
            Index to use to generate shade collection

        Returns
        -------
        collection : :py:class:`~pvfactors.geometry.base.ShadeCollection`
        """
        list_surfaces = [ts_surf.at(idx) for ts_surf in self._list_ts_surfaces
                         if not ts_surf.at(idx).is_empty]
        return ShadeCollection(list_surfaces, shaded=self.shaded)


class TsSurface(object):
    """Timeseries surface class: vectorized representation of PV surface
    geometries."""

    def __init__(self, coords, n_vector=None, param_names=None, index=None,
                 shaded=False):
        """Initialize timeseries surface using timeseries coordinates.

        Parameters
        ----------
        coords : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries coordinates of full segment
        index : int, optional
            Index of segment (Default = None)
        n_vector : np.ndarray, optional
            Timeseries normal vectors of the side (Default = None)
        index : int, optional
            Index of the timeseries surfaces (Default = None)
        shaded : bool, optional
            Is the surface shaded or not (Default = False)
        """
        # TODO: ts surfaces should have a shaded attribute
        self.coords = coords
        self.param_names = [] if param_names is None else param_names
        # TODO: the following should probably be turned into properties,
        # because if the coords change, they won't be altered. But speed...
        self.n_vector = n_vector
        self.params = dict.fromkeys(self.param_names)
        self.index = index
        self.shaded = shaded

    def at(self, idx):
        """Generate a PV segment geometry for the desired index.

        Parameters
        ----------
        idx : int
            Index to use to generate PV segment geometry

        Returns
        -------
        segment : :py:class:`~pvfactors.geometry.base.PVSurface` \
        or :py:class:`~shapely.geometry.GeometryCollection`
            The returned object will be an empty geometry if its length is
            really small, otherwise it will be a PV surface geometry
        """
        if self.length[idx] < DISTANCE_TOLERANCE:
            # return an empty geometry
            return GeometryCollection()
        else:
            # Get normal vector at idx
            n_vector = (self.n_vector[:, idx] if self.n_vector is not None
                        else None)
            # Get params at idx
            # TODO: should find faster solution
            params = _get_params_at_idx(idx, self.params)
            # Return a pv surface geometry with given params
            return PVSurface(self.coords.at(idx), shaded=self.shaded,
                             index=self.index, normal_vector=n_vector,
                             param_names=self.param_names,
                             params=params)

    def plot_at_idx(self, idx, ax, color):
        """Plot timeseries PV row at a certain index, only if it's not
        too small.

        Parameters
        ----------
        idx : int
            Index to use to plot timeseries PV surface
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting
        color_shaded : str, optional
            Color to use for plotting the PV surface
        """
        if self.length[idx] > DISTANCE_TOLERANCE:
            self.at(idx).plot(ax, color=color)

    @property
    def b1(self):
        """Timeseries coordinates of first boundary point"""
        return self.coords.b1

    @property
    def b2(self):
        """Timeseries coordinates of second boundary point"""
        return self.coords.b2

    @property
    def centroid(self):
        """Timeseries point coordinates of the surface's centroid"""
        return self.coords.centroid

    def get_param(self, param):
        """Get timeseries parameter values of surface

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        np.ndarray
            Timeseries parameter values
        """
        return self.params[param]

    def update_params(self, new_dict):
        """Update timeseries surface parameters.

        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surface
        """
        self.params.update(new_dict)

    @property
    def length(self):
        """Timeseries length of the surface"""
        return self.coords.length

    @property
    def highest_point(self):
        """Timeseries point coordinates of highest point of surface"""
        return self.coords.highest_point

    @property
    def lowest_point(self):
        """Timeseries point coordinates of lowest point of surface"""
        return self.coords.lowest_point

    @property
    def u_vector(self):
        """Vector orthogonal to the surface's normal vector"""
        u_vector = (None if self.n_vector is None else
                    np.array([-self.n_vector[1, :], self.n_vector[0, :]]))
        return u_vector

    @property
    def is_empty(self):
        """Check if surface is "empty" by checking if its length is always
        zero"""
        return np.nansum(self.length) < DISTANCE_TOLERANCE


class TsLineCoords(object):
    """Timeseries line coordinates class: will provide a helpful shapely-like
    API to invoke timeseries coordinates."""

    def __init__(self, b1_ts_coords, b2_ts_coords, coords=None):
        """Initialize timeseries line coordinates using the timeseries
        coordinates of its boundaries.

        Parameters
        ----------
        b1_ts_coords : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of first boundary point
        b2_ts_coords : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of second boundary point
        coords : np.ndarray, optional
            Timeseries coordinates as numpy array
        """
        self.b1 = b1_ts_coords
        self.b2 = b2_ts_coords

    def at(self, idx):
        """Get coordinates at a given index

        Parameters
        ----------
        idx : int
            Index to use to get coordinates
        """
        return self.as_array[:, :, idx]

    @classmethod
    def from_array(cls, coords_array):
        """Create timeseries line coordinates from numpy array of coordinates.

        Parameters
        ----------
        coords_array : np.ndarray
            Numpy array of coordinates.
        """
        b1 = TsPointCoords.from_array(coords_array[0, :, :])
        b2 = TsPointCoords.from_array(coords_array[1, :, :])
        return cls(b1, b2)

    @property
    def length(self):
        """Timeseries length of the line."""
        return np.sqrt((self.b2.y - self.b1.y)**2
                       + (self.b2.x - self.b1.x)**2)

    @property
    def as_array(self):
        """Timeseries line coordinates as numpy array"""
        return np.array([[self.b1.x, self.b1.y], [self.b2.x, self.b2.y]])

    @property
    def centroid(self):
        """Timeseries point coordinates of the line coordinates"""
        dy = self.b2.y - self.b1.y
        dx = self.b2.x - self.b1.x
        return TsPointCoords(self.b1.x + 0.5 * dx, self.b1.y + 0.5 * dy)

    @property
    def highest_point(self):
        """Timeseries point coordinates of highest point of timeseries
        line coords"""
        is_b1_highest = self.b1.y >= self.b2.y
        x = np.where(is_b1_highest, self.b1.x, self.b2.x)
        y = np.where(is_b1_highest, self.b1.y, self.b2.y)
        return TsPointCoords(x, y)

    @property
    def lowest_point(self):
        """Timeseries point coordinates of lowest point of timeseries
        line coords"""
        is_b1_highest = self.b1.y >= self.b2.y
        x = np.where(is_b1_highest, self.b2.x, self.b1.x)
        y = np.where(is_b1_highest, self.b2.y, self.b1.y)
        return TsPointCoords(x, y)

    def __repr__(self):
        """Use the numpy array representation of the coords"""
        return str(self.as_array)


class TsPointCoords(object):
    """Timeseries point coordinates: provides a shapely-like API for timeseries
    point coordinates."""

    def __init__(self, x, y):
        """Initialize timeseries point coordinates using numpy array of coords.

        Parameters
        ----------
        x : np.ndarray
            Timeseries x coordinates
        y : np.ndarray
            Timeseries y coordinates
        """
        self.x = x
        self.y = y

    def at(self, idx):
        """Get coordinates at a given index

        Parameters
        ----------
        idx : int
            Index to use to get coordinates
        """
        return self.as_array[:, idx]

    @property
    def as_array(self):
        """Timeseries point coordinates as numpy array"""
        return np.array([self.x, self.y])

    @classmethod
    def from_array(cls, coords_array):
        """Create timeseries point coords from numpy array of coordinates.

        Parameters
        ----------
        coords_array : np.ndarray
            Numpy array of coordinates.
        """
        return cls(coords_array[0, :], coords_array[1, :])

    def __repr__(self):
        """Use the numpy array representation of the point"""
        return str(self.as_array)


def _get_params_at_idx(idx, params_dict):
    """Get the parameter values at given index. Return the whole parameter
    when it's None, a scalar, or a dictionary

    Parameters
    ----------
    idx : int
        Index at which we want the parameter values
    params_dict : dict
        Dictionary of parameters

    Returns
    -------
    Parameter value at index
    """
    if params_dict is None:
        return None
    else:
        return {k: (val if (val is None) or np.isscalar(val)
                    or isinstance(val, dict) else val[idx])
                for k, val in params_dict.items()}



# -*- coding: utf-8 -*-

from pvfactors.version import __version__
import logging
logging.basicConfig()

try:
    from shapely.geos import lgeos
except OSError as err:
    # https://github.com/SunPower/pvfactors/issues/109
    msg = (
        "pvfactors encountered an error when importing the shapely package. "
        "This often happens when a binary dependency is missing because "
        "shapely was installed from PyPI using pip. Try reinstalling shapely "
        "from another source like conda-forge with "
        "`conda install -c conda-forge shapely`, or alternatively from "
        "Christoph Gohlke's website if you're on Windows: "
        "https://www.lfd.uci.edu/~gohlke/pythonlibs/#shapely"
    )
    err.strerror += "; " + msg
    raise err


class PVFactorsError(Exception):
    pass



"""Base classes for pvfactors geometry subpackage."""

import numpy as np
from pvfactors import PVFactorsError
from pvfactors.config import (
    DEFAULT_NORMAL_VEC, COLOR_DIC, DISTANCE_TOLERANCE, PLOT_FONTSIZE,
    ALPHA_TEXT, MAX_X_GROUND)
from pvfactors.geometry.plot import plot_coords, plot_bounds, plot_line
from pvfactors.geometry.utils import \
    is_collinear, check_collinear, are_2d_vecs_collinear, difference, contains
from shapely.geometry import GeometryCollection, LineString
from shapely.geometry.collection import geos_geometrycollection_from_py
from shapely.ops import linemerge
from pvlib.tools import cosd, sind


def _check_uniform_shading(list_elements):
    """Check that all :py:class:`~pvfactors.geometry.base.PVSurface` objects in
    list have uniform shading

    Parameters
    ----------
    list_elements : list of :py:class:`~pvfactors.geometry.base.PVSurface`

    Raises
    ------
    PVFactorsError
        if all elements don't have the same shading flag
    """
    shaded = None
    for element in list_elements:
        if shaded is None:
            shaded = element.shaded
        else:
            is_uniform = shaded == element.shaded
            if not is_uniform:
                msg = "All elements should have same shading"
                raise PVFactorsError(msg)


def _coords_from_center_tilt_length(xy_center, tilt, length,
                                    surface_azimuth, axis_azimuth):
    """Calculate ``shapely`` :py:class:`LineString` coordinates from
    center coords, surface angles and length of line.
    The axis azimuth indicates the axis of rotation of the pvrows (if single-
    axis trackers). In the 2D plane, the axis of rotation will be the vector
    normal to that 2D plane and going into the 2D plane (when plotting it).
    The surface azimuth should always be 90 degrees away from the axis azimuth,
    either in the positive or negative direction.
    For instance, a single axis trk with axis azimuth = 0 deg (North), will
    have surface azimuth values equal to 90 deg (East) or 270 deg (West).
    Tilt angles need to always be positive. Given the axis azimuth and surface
    azimuth, a rotation angle will be derived. Positive rotation angles will
    indicate pvrows pointing to the left, and negative rotation angles will
    indicate pvrows pointing to the right (no matter what the the axis azimuth
    is).
    All of these conventions are necessary to make sure that no matter what
    the tilt and surface angles are, we can still identify correctly
    the same pv rows: the leftmost PV row will have index 0, and the rightmost
    will have index -1.

    Parameters
    ----------
    xy_center : tuple
        x, y coordinates of center point of desired linestring
    tilt : float or np.ndarray
        Surface tilt angles desired [deg]. Values should all be positive.
    length : float
        desired length of linestring [m]
    surface_azimuth : float or np.ndarray
        Surface azimuth angles of PV surface [deg]
    axis_azimuth : float
        Axis azimuth of the PV surface, i.e. direction of axis of rotation
        [deg]

    Returns
    -------
    list
        List of linestring coordinates obtained from inputs (could be vectors)
        in the form of [[x1, y1], [x2, y2]], where xi and yi could be arrays
        or scalar values.
    """
    # PV row params
    x_center, y_center = xy_center
    radius = length / 2.
    # Get rotation
    rotation = _get_rotation_from_tilt_azimuth(surface_azimuth, axis_azimuth,
                                               tilt)
    # Calculate coords
    x1 = radius * cosd(rotation + 180.) + x_center
    y1 = radius * sind(rotation + 180.) + y_center
    x2 = radius * cosd(rotation) + x_center
    y2 = radius * sind(rotation) + y_center

    return [[x1, y1], [x2, y2]]


def _get_rotation_from_tilt_azimuth(surface_azimuth, axis_azimuth, tilt):
    """Calculate the rotation angle using surface azimuth, axis azimuth,
    and surface tilt angles. While surface tilt angles need to always be
    positive, rotation angles can be negative.
    In pvfactors, positive rotation angles will indicate pvrows pointing to the
    left, and negative rotation angles will indicate pvrows pointing to the
    right (no matter what the the axis azimuth is).

    Parameters
    ----------
    tilt : float or np.ndarray
        Surface tilt angles desired [deg]. Values should all be positive.
    surface_azimuth : float or np.ndarray
        Surface azimuth angles of PV surface [deg]
    axis_azimuth : float
        Axis azimuth of the PV surface, i.e. direction of axis of rotation
        [deg]

    Returns
    -------
    float or np.ndarray
        Calculated rotation angle(s) in [deg]
    """

    # Calculate rotation of PV row (signed tilt angle)
    is_pointing_right = ((surface_azimuth - axis_azimuth) % 360.) > 180.
    rotation = np.where(is_pointing_right, tilt, -tilt)

    return rotation


def _get_solar_2d_vectors(solar_zenith, solar_azimuth, axis_azimuth):
    """Projection of 3d solar vector onto the cross section of the systems:
    which is the 2D plane we are considering.
    This is needed to calculate shadows.
    Remember that the 2D plane is such that the direction of the torque
    tube vector (or rotation axis) goes into (and normal to) the 2D plane,
    such that positive rotation angles will have the PV surfaces tilted to the
    LEFT and vice versa.

    Parameters
    ----------
    solar_zenith : float or numpy array
        Solar zenith angle [deg]
    solar_azimuth : float or numpy array
        Solar azimuth angle [deg]
    axis_azimuth : float
        Axis azimuth of the PV surface, i.e. direction of axis of rotation
        [deg]

    Returns
    -------
    solar_2d_vector : numpy array
        Two vector components of the solar vector in the 2D plane, with the
        form [x, y], where x and y can be arrays
    """
    solar_2d_vector = np.array([
        # a drawing really helps understand the following
        sind(solar_zenith) * cosd(solar_azimuth - axis_azimuth - 90.),
        cosd(solar_zenith)])

    return solar_2d_vector


class BaseSurface(LineString):
    """Base surfaces will be extensions of :py:class:`LineString` classes,
    but adding an orientation to it (normal vector).
    So two surfaces could use the same linestring, but have opposite
    orientations."""

    def __init__(self, coords, normal_vector=None, index=None,
                 param_names=None, params=None):
        """Create a surface using linestring coordinates.
        Normal vector can have two directions for a given LineString,
        so the user can provide it in order to be specific,
        otherwise it will be automatically
        calculated, but then the surface won't know if it was supposed to be
        pointing "up" or "down". If the surface is empty, the normal vector
        will take the default value.

        Parameters
        ----------
        coords : list
            List of linestring coordinates for the surface
        normal_vector : list, optional
            Normal vector for the surface (Default = None, so will be
            calculated)
        index : int, optional
            Surface index (Default = None)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = None)
        params : dict, optional
            Surface float parameters (Default = None)
        """

        param_names = [] if param_names is None else param_names
        super(BaseSurface, self).__init__(coords)
        if normal_vector is None:
            self.n_vector = self._calculate_n_vector()
        else:
            self.n_vector = np.array(normal_vector)
        self.index = index
        self.param_names = param_names
        self.params = params if params is not None \
            else dict.fromkeys(self.param_names)

    def _calculate_n_vector(self):
        """Calculate normal vector of the surface, if surface is not empty"""
        if not self.is_empty:
            b1, b2 = self.boundary
            dx = b2.x - b1.x
            dy = b2.y - b1.y
            return np.array([-dy, dx])
        else:
            return DEFAULT_NORMAL_VEC

    def plot(self, ax, color=None, with_index=False):
        """Plot the surface on the given axes.

        Parameters
        ----------
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting
        color : str, optional
            Color to use for plotting the surface (Default = None)
        with_index : bool
            Flag to annotate surfaces with their indices (Default = False)
        """
        plot_coords(ax, self)
        plot_bounds(ax, self)
        plot_line(ax, self, color)
        if with_index:
            # Prepare text location
            v = self.n_vector
            v_norm = v / np.linalg.norm(v)
            centroid = self.centroid
            alpha = ALPHA_TEXT
            x = centroid.x + alpha * v_norm[0]
            y = centroid.y + alpha * v_norm[1]
            # Add text
            # FIXME: hack to get a nice plot in jupyter notebook
            if np.abs(x) < MAX_X_GROUND / 2.:
                ax.text(x, y, '{}'.format(self.index),
                        verticalalignment='center',
                        horizontalalignment='center')

    def difference(self, linestring):
        """Calculate remaining surface after removing part belonging from
        provided linestring,

        Parameters
        ----------
        linestring : :py:class:`shapely.geometry.LineString`
            Line string to remove from surface

        Returns
        -------
        :py:class:`shapely.geometry.LineString`
           Resulting difference of current surface minus given linestring
        """
        return difference(self, linestring)

    def get_param(self, param):
        """Get parameter value from surface.

        Parameters
        ----------
        param : str
            Surface parameter to return

        Returns
        -------
        Parameter value to return

        Raises
        ------
        KeyError
            if parameter name not in the surface parameters
        """
        return self.params[param]

    def update_params(self, new_dict):
        """Update surface parameters.

        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surface
        """
        self.params.update(new_dict)


class PVSurface(BaseSurface):
    """PV surfaces inherit from
    :py:class:`~pvfactors.geometry.base.BaseSurface`. The only difference is
    that PV surfaces have a ``shaded`` attribute.
    """

    def __init__(self, coords=None, normal_vector=None, shaded=False,
                 index=None, param_names=None, params=None):
        """Initialize PV surface.

        Parameters
        ----------
        coords : list, optional
            List of linestring coordinates for the surface
        normal_vector : list, optional
            Normal vector for the surface (Default = None, so will be
            calculated)
        shaded : bool, optional
            Flag telling if surface is shaded or not (Default = False)
        index : int, optional
            Surface index (Default = None)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = None)
        params : dict, optional
            Surface float parameters (Default = None)
        """

        param_names = [] if param_names is None else param_names
        super(PVSurface, self).__init__(coords, normal_vector, index=index,
                                        param_names=param_names, params=params)
        self.shaded = shaded


class ShadeCollection(GeometryCollection):
    """A group of :py:class:`~pvfactors.geometry.base.PVSurface`
    objects that all have the same shading status. The PV surfaces are not
    necessarily contiguous or collinear."""

    def __init__(self, list_surfaces=None, shaded=None, param_names=None):
        """Initialize shade collection.

        Parameters
        ----------
        list_surfaces : list, optional
            List of :py:class:`~pvfactors.geometry.base.PVSurface` object
            (Default = None)
        shaded : bool, optional
            Shading status of the collection. If not specified, will be derived
            from list of surfaces (Default = None)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = None)

        """
        list_surfaces = [] if list_surfaces is None else list_surfaces
        param_names = [] if param_names is None else param_names
        _check_uniform_shading(list_surfaces)
        self.list_surfaces = list_surfaces
        self.shaded = self._get_shading(shaded)
        self.is_collinear = is_collinear(list_surfaces)
        self.param_names = param_names
        super(ShadeCollection, self).__init__(list_surfaces)

    def _get_shading(self, shaded):
        """Get the surface shading from the provided list of pv surfaces.

        Parameters
        ----------
        shaded : bool
            Shading flag passed during initialization

        Returns
        -------
        bool
            Shading status of the collection
        """
        if len(self.list_surfaces):
            return self.list_surfaces[0].shaded
        else:
            return shaded

    def plot(self, ax, color=None, with_index=False):
        """Plot the surfaces in the shade collection.

        Parameters
        ----------
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting
        color : str, optional
            Color to use for plotting the surface (Default = None)
        with_index : bool
            Flag to annotate surfaces with their indices (Default = False)
        """
        for surface in self.list_surfaces:
            surface.plot(ax, color=color, with_index=with_index)

    def add_linestring(self, linestring, normal_vector=None):
        """Add PV surface to the collection using a linestring

        Parameters
        ----------
        linestring : :py:class:`shapely.geometry.LineString`
            Linestring to use to add a PV surface to the collection
        normal_vector : list, optional
            Normal vector to use for the PV surface to create (Default = None,
            will try to get it from collection)
        """
        if normal_vector is None:
            normal_vector = self.n_vector
        surf = PVSurface(coords=linestring.coords,
                         normal_vector=normal_vector, shaded=self.shaded,
                         param_names=self.param_names)
        self.add_pvsurface(surf)

    def add_pvsurface(self, pvsurface):
        """Add PV surface to the collection.

        Parameters
        ----------
        pvsurface : :py:class:`~pvfactors.geometry.base.PVSurface`
            PV Surface to add to collection
        """
        self.list_surfaces.append(pvsurface)
        self.is_collinear = is_collinear(self.list_surfaces)
        super(ShadeCollection, self).__init__(self.list_surfaces)

    def remove_linestring(self, linestring):
        """Remove linestring from shade collection.
        The method will rearrange the PV surfaces to make it work.

        Parameters
        ----------
        linestring : :py:class:`shapely.geometry.LineString`
            Line string to remove from the collection (by differencing)
        """
        new_list_surfaces = []
        for surface in self.list_surfaces:
            # Need to use buffer for intersects bc of floating point precision
            # errors in shapely
            if surface.buffer(DISTANCE_TOLERANCE).intersects(linestring):
                difference = surface.difference(linestring)
                # We want to make sure we can iterate on it, as
                # ``difference`` can be a multi-part geometry or not
                if not hasattr(difference, '__iter__'):
                    difference = [difference]
                for new_geom in difference:
                    if not new_geom.is_empty:
                        new_surface = PVSurface(
                            new_geom.coords, normal_vector=surface.n_vector,
                            shaded=surface.shaded,
                            param_names=surface.param_names)
                        new_list_surfaces.append(new_surface)
            else:
                new_list_surfaces.append(surface)

        self.list_surfaces = new_list_surfaces
        # Force update, even if list is empty
        self.update_geom_collection(self.list_surfaces)

    def update_geom_collection(self, list_surfaces):
        """Force update of geometry collection, even if list is empty
        https://github.com/Toblerity/Shapely/blob/master/shapely/geometry/collection.py#L42

        Parameters
        ----------
        list_surfaces : list of :py:class:`~pvfactors.geometry.base.PVSurface`
            New list of PV surfaces to update the shade collection in place
        """
        self._geom, self._ndim = geos_geometrycollection_from_py(list_surfaces)

    def merge_surfaces(self):
        """Merge all surfaces in the shade collection into one contiguous
        surface, even if they're not contiguous, by using bounds."""
        if len(self.list_surfaces) > 1:
            merged_lines = linemerge(self.list_surfaces)
            minx, miny, maxx, maxy = merged_lines.bounds
            surf_1 = self.list_surfaces[0]
            new_pvsurf = PVSurface(
                coords=[(minx, miny), (maxx, maxy)],
                shaded=self.shaded, normal_vector=surf_1.n_vector,
                param_names=surf_1.param_names)
            self.list_surfaces = [new_pvsurf]
            self.update_geom_collection(self.list_surfaces)

    def cut_at_point(self, point):
        """Cut collection at point if the collection contains it.

        Parameters
        ----------
        point : :py:class:`shapely.geometry.Point`
            Point where to cut collection geometry, if the latter contains the
            former
        """
        for idx, surface in enumerate(self.list_surfaces):
            if contains(surface, point):
                # Make sure that not hitting a boundary
                b1, b2 = surface.boundary
                not_hitting_b1 = b1.distance(point) > DISTANCE_TOLERANCE
                not_hitting_b2 = b2.distance(point) > DISTANCE_TOLERANCE
                if not_hitting_b1 and not_hitting_b2:
                    coords_1 = [b1, point]
                    coords_2 = [point, b2]
                    # TODO: not sure what to do about index yet
                    new_surf_1 = PVSurface(
                        coords_1, normal_vector=surface.n_vector,
                        shaded=surface.shaded,
                        param_names=surface.param_names)
                    new_surf_2 = PVSurface(
                        coords_2, normal_vector=surface.n_vector,
                        shaded=surface.shaded,
                        param_names=surface.param_names)
                    # Now update collection
                    self.list_surfaces[idx] = new_surf_1
                    self.list_surfaces.append(new_surf_2)
                    self.update_geom_collection(self.list_surfaces)
                    # No need to continue the loop
                    break

    def get_param_weighted(self, param):
        """Get the parameter from the collection's surfaces, after weighting
        by surface length.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        float
            Weighted parameter value
        """
        value = self.get_param_ww(param) / self.length
        return value

    def get_param_ww(self, param):
        """Get the parameter from the collection's surfaces with weight, i.e.
        after multiplying by the surface lengths.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        float
            Parameter value multiplied by weights

        Raises
        ------
        KeyError
            if parameter name not in a surface parameters
        """
        value = 0
        for surf in self.list_surfaces:
            value += surf.get_param(param) * surf.length
        return value

    def update_params(self, new_dict):
        """Update surface parameters in the collection.

        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surface
        """
        for surf in self.list_surfaces:
            surf.update_params(new_dict)

    @property
    def n_vector(self):
        """Unique normal vector of the shade collection, if it exists."""
        if not self.is_collinear:
            msg = "Cannot request n_vector if all elements not collinear"
            raise PVFactorsError(msg)
        if len(self.list_surfaces):
            return self.list_surfaces[0].n_vector
        else:
            return DEFAULT_NORMAL_VEC

    @property
    def n_surfaces(self):
        """Number of surfaces in collection."""
        return len(self.list_surfaces)

    @property
    def surface_indices(self):
        """Indices of the surfaces in the collection."""
        return [surf.index for surf in self.list_surfaces]

    @classmethod
    def from_linestring_coords(cls, coords, shaded, normal_vector=None,
                               param_names=None):
        """Create a shade collection with a single PV surface.

        Parameters
        ----------
        coords : list
            List of linestring coordinates for the surface
        shaded : bool
            Shading status desired for the collection
        normal_vector : list, optional
            Normal vector for the surface (Default = None)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = None)
        """
        surf = PVSurface(coords=coords, normal_vector=normal_vector,
                         shaded=shaded, param_names=param_names)
        return cls([surf], shaded=shaded, param_names=param_names)


class PVSegment(GeometryCollection):
    """A PV segment will be a collection of 2 collinear and contiguous
    shade collections, a shaded one and an illuminated one. It inherits from
    :py:class:`shapely.geometry.GeometryCollection`  so that users can still
    call basic geometrical methods and properties on it, eg call length, etc.
    """

    def __init__(self, illum_collection=ShadeCollection(shaded=False),
                 shaded_collection=ShadeCollection(shaded=True), index=None):
        """Initialize PV segment.

        Parameters
        ----------
        illum_collection : \
        :py:class:`~pvfactors.geometry.base.ShadeCollection`, optional
            Illuminated collection of the PV segment (Default = empty shade
            collection with no shading)
        shaded_collection : \
        :py:class:`~pvfactors.geometry.base.ShadeCollection`, optional
            Shaded collection of the PV segment (Default = empty shade
            collection with shading)
        index : int, optional
            Index of the PV segment (Default = None)
        """
        assert shaded_collection.shaded, "surface should be shaded"
        assert not illum_collection.shaded, "surface should not be shaded"
        self._check_collinear(illum_collection, shaded_collection)
        self._shaded_collection = shaded_collection
        self._illum_collection = illum_collection
        self.index = index
        self._all_surfaces = None
        super(PVSegment, self).__init__([self._shaded_collection,
                                         self._illum_collection])

    def _check_collinear(self, illum_collection, shaded_collection):
        """Check that all the surfaces in the PV segment are collinear.

        Parameters
        ----------
        illum_collection :
        :py:class:`~pvfactors.geometry.base.ShadeCollection`, optional
            Illuminated collection
        shaded_collection :
        :py:class:`~pvfactors.geometry.base.ShadeCollection`, optional
            Shaded collection

        Raises
        ------
        PVFactorsError
            If all the surfaces are not collinear
        """
        assert illum_collection.is_collinear
        assert shaded_collection.is_collinear
        # Check that if none or all of the collection is empty, n_vectors are
        # equal
        if (not illum_collection.is_empty) \
           and (not shaded_collection.is_empty):
            n_vec_ill = illum_collection.n_vector
            n_vec_shaded = shaded_collection.n_vector
            assert are_2d_vecs_collinear(n_vec_ill, n_vec_shaded)

    def plot(self, ax, color_shaded=COLOR_DIC['pvrow_shaded'],
             color_illum=COLOR_DIC['pvrow_illum'], with_index=False):
        """Plot the surfaces in the PV Segment.

        Parameters
        ----------
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting
        color_shaded : str, optional
            Color to use for plotting the shaded surfaces (Default =
            COLOR_DIC['pvrow_shaded'])
        color_shaded : str, optional
            Color to use for plotting the illuminated surfaces (Default =
            COLOR_DIC['pvrow_illum'])
        with_index : bool
            Flag to annotate surfaces with their indices (Default = False)
        """
        self._shaded_collection.plot(ax, color=color_shaded,
                                     with_index=with_index)
        self._illum_collection.plot(ax, color=color_illum,
                                    with_index=with_index)

    def cast_shadow(self, linestring):
        """Cast shadow on PV segment using linestring: will rearrange the
        PV surfaces between the shaded and illuminated collections of the
        segment

        Parameters
        ----------
        linestring : :py:class:`shapely.geometry.LineString`
            Linestring casting a shadow on the PV segment
        """
        # Using a buffer may slow things down, but it's quite crucial
        # in order for shapely to get the intersection accurately see:
        # https://stackoverflow.com/questions/28028910/how-to-deal-with-rounding-errors-in-shapely
        intersection = (self._illum_collection.buffer(DISTANCE_TOLERANCE)
                        .intersection(linestring))
        if not intersection.is_empty:
            # Split up only if interesects the illuminated collection
            # print(intersection)
            self._shaded_collection.add_linestring(intersection,
                                                   normal_vector=self.n_vector)
            # print(self._shaded_collection.length)
            self._illum_collection.remove_linestring(intersection)
            # print(self._illum_collection.length)
            super(PVSegment, self).__init__([self._shaded_collection,
                                             self._illum_collection])

    def cut_at_point(self, point):
        """Cut PV segment at point if the segment contains it.

        Parameters
        ----------
        point : :py:class:`shapely.geometry.Point`
            Point where to cut collection geometry, if the latter contains the
            former
        """
        if contains(self, point):
            if contains(self._illum_collection, point):
                self._illum_collection.cut_at_point(point)
            else:
                self._shaded_collection.cut_at_point(point)

    def get_param_weighted(self, param):
        """Get the parameter from the segment's surfaces, after weighting
        by surface length.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        float
            Weighted parameter value
        """
        value = self.get_param_ww(param) / self.length
        return value

    def get_param_ww(self, param):
        """Get the parameter from the segment's surfaces with weight, i.e.
        after multiplying by the surface lengths.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        float
            Parameter value multiplied by weights

        Raises
        ------
        KeyError
            if parameter name not in a surface parameters
        """
        value = 0
        value += self._shaded_collection.get_param_ww(param)
        value += self._illum_collection.get_param_ww(param)
        return value

    def update_params(self, new_dict):
        """Update surface parameters in the collection.

        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surfaces
        """
        self._shaded_collection.update_params(new_dict)
        self._illum_collection.update_params(new_dict)

    @property
    def n_vector(self):
        """Since shaded and illum surfaces are supposed to be collinear,
        this should return either surfaces' normal vector. If both empty,
        return Default value for normal vector."""
        if not self.illum_collection.is_empty:
            return self.illum_collection.n_vector
        elif not self.shaded_collection.is_empty:
            return self.shaded_collection.n_vector
        else:
            return DEFAULT_NORMAL_VEC

    @property
    def n_surfaces(self):
        """Number of surfaces in collection."""
        n_surfaces = self._illum_collection.n_surfaces \
            + self._shaded_collection.n_surfaces
        return n_surfaces

    @property
    def surface_indices(self):
        """Indices of the surfaces in the PV segment."""
        list_indices = []
        list_indices += self._illum_collection.surface_indices
        list_indices += self._shaded_collection.surface_indices
        return list_indices

    @classmethod
    def from_linestring_coords(cls, coords, shaded=False, normal_vector=None,
                               index=None, param_names=None):
        """Create a PV segment with a single PV surface.

        Parameters
        ----------
        coords : list
            List of linestring coordinates for the surface
        shaded : bool, optional
            Shading status desired for the resulting PV surface
            (Default = False)
        normal_vector : list, optional
            Normal vector for the surface (Default = None)
        index : int, optional
            Index of the segment (Default = None)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = None)
        """
        col = ShadeCollection.from_linestring_coords(
            coords, shaded=shaded, normal_vector=normal_vector,
            param_names=param_names)
        # Realized that needed to instantiate other_col, otherwise could
        # end up with shared collection among different PV segments
        other_col = ShadeCollection(list_surfaces=[], shaded=not shaded,
                                    param_names=param_names)
        if shaded:
            return cls(illum_collection=other_col,
                       shaded_collection=col, index=index)
        else:
            return cls(illum_collection=col,
                       shaded_collection=other_col, index=index)

    @property
    def shaded_collection(self):
        """Shaded collection of the PV segment"""
        return self._shaded_collection

    @shaded_collection.setter
    def shaded_collection(self, new_collection):
        """Set shaded collection of the PV segment with new one.

        Parameters
        ----------
        new_collection : :py:class:`pvfactors.geometry.base.ShadeCollection`
            New collection to use for update
        """
        assert new_collection.shaded, "surface should be shaded"
        self._shaded_collection = new_collection
        super(PVSegment, self).__init__([self._shaded_collection,
                                         self._illum_collection])

    @shaded_collection.deleter
    def shaded_collection(self):
        """Delete shaded collection of PV segment and replace with empty one.
        """
        self._shaded_collection = ShadeCollection(shaded=True)
        super(PVSegment, self).__init__([self._shaded_collection,
                                         self._illum_collection])

    @property
    def illum_collection(self):
        """Illuminated collection of the PV segment."""
        return self._illum_collection

    @illum_collection.setter
    def illum_collection(self, new_collection):
        """Set illuminated collection of the PV segment with new one.

        Parameters
        ----------
        new_collection : :py:class:`pvfactors.geometry.base.ShadeCollection`
            New collection to use for update
        """
        assert not new_collection.shaded, "surface should not be shaded"
        self._illum_collection = new_collection
        super(PVSegment, self).__init__([self._shaded_collection,
                                         self._illum_collection])

    @illum_collection.deleter
    def illum_collection(self):
        """Delete illuminated collection of PV segment and replace with empty
        one."""
        self._illum_collection = ShadeCollection(shaded=False)
        super(PVSegment, self).__init__([self._shaded_collection,
                                         self._illum_collection])

    @property
    def shaded_length(self):
        """Length of the shaded collection of the PV segment.

        Returns
        -------
        float
            Length of the shaded collection
        """
        return self._shaded_collection.length

    @property
    def all_surfaces(self):
        """List of all the :py:class:`pvfactors.geometry.base.PVSurface`

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
            PV surfaces in the PV segment
        """
        if self._all_surfaces is None:
            self._all_surfaces = []
            self._all_surfaces += self._illum_collection.list_surfaces
            self._all_surfaces += self._shaded_collection.list_surfaces
        return self._all_surfaces


class BaseSide(GeometryCollection):
    """A side represents a fixed collection of PV segments objects that should
    all be collinear, with the same normal vector"""

    def __init__(self, list_segments=None):
        """Create a side geometry.

        Parameters
        ----------
        list_segments : list of :py:class:`~pvfactors.geometry.base.PVSegment`, optional
            List of PV segments for side (Default = None)
        """
        list_segments = [] if list_segments is None else list_segments
        check_collinear(list_segments)
        self.list_segments = tuple(list_segments)
        self._all_surfaces = None
        super(BaseSide, self).__init__(list_segments)

    @classmethod
    def from_linestring_coords(cls, coords, shaded=False, normal_vector=None,
                               index=None, n_segments=1, param_names=None):
        """Create a Side with a single PV surface, or multiple discretized
        identical ones.

        Parameters
        ----------
        coords : list
            List of linestring coordinates for the surface
        shaded : bool, optional
            Shading status desired for the resulting PV surface
            (Default = False)
        normal_vector : list, optional
            Normal vector for the surface (Default = None)
        index : int, optional
            Index of the segments (Default = None)
        n_segments : int, optional
            Number of same-length segments to use (Default = 1)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = None)
        """
        if n_segments == 1:
            list_pvsegments = [PVSegment.from_linestring_coords(
                coords, shaded=shaded, normal_vector=normal_vector,
                index=index, param_names=param_names)]
        else:
            # Discretize coords and create segments accordingly
            linestring = LineString(coords)
            fractions = np.linspace(0., 1., num=n_segments + 1)
            list_points = [linestring.interpolate(fraction, normalized=True)
                           for fraction in fractions]
            list_pvsegments = []
            for idx in range(n_segments):
                new_coords = list_points[idx:idx + 2]
                # TODO: not clear what to do with the index here
                pvsegment = PVSegment.from_linestring_coords(
                    new_coords, shaded=shaded, normal_vector=normal_vector,
                    index=index, param_names=param_names)
                list_pvsegments.append(pvsegment)
        return cls(list_segments=list_pvsegments)

    @property
    def n_vector(self):
        """Normal vector of the Side."""
        if len(self.list_segments):
            return self.list_segments[0].n_vector
        else:
            return DEFAULT_NORMAL_VEC

    @property
    def shaded_length(self):
        """Shaded length of the Side."""
        shaded_length = 0.
        for segment in self.list_segments:
            shaded_length += segment.shaded_length
        return shaded_length

    @property
    def n_surfaces(self):
        """Number of surfaces in the Side object."""
        n_surfaces = 0
        for segment in self.list_segments:
            n_surfaces += segment.n_surfaces
        return n_surfaces

    @property
    def all_surfaces(self):
        """List of all surfaces in the Side object."""
        if self._all_surfaces is None:
            self._all_surfaces = []
            for segment in self.list_segments:
                self._all_surfaces += segment.all_surfaces
        return self._all_surfaces

    @property
    def surface_indices(self):
        """List of all surface indices in the Side object."""
        list_indices = []
        for seg in self.list_segments:
            list_indices += seg.surface_indices
        return list_indices

    def plot(self, ax, color_shaded=COLOR_DIC['pvrow_shaded'],
             color_illum=COLOR_DIC['pvrow_illum'], with_index=False):
        """Plot the surfaces in the Side object.

        Parameters
        ----------
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting
        color_shaded : str, optional
            Color to use for plotting the shaded surfaces (Default =
            COLOR_DIC['pvrow_shaded'])
        color_shaded : str, optional
            Color to use for plotting the illuminated surfaces (Default =
            COLOR_DIC['pvrow_illum'])
        with_index : bool
            Flag to annotate surfaces with their indices (Default = False)
        """
        for segment in self.list_segments:
            segment.plot(ax, color_shaded=color_shaded,
                         color_illum=color_illum, with_index=with_index)

    def cast_shadow(self, linestring):
        """Cast shadow on Side using linestring: will rearrange the
        PV surfaces between the shaded and illuminated collections of the
        segments.

        Parameters
        ----------
        linestring : :py:class:`shapely.geometry.LineString`
            Linestring casting a shadow on the Side object
        """
        for segment in self.list_segments:
            segment.cast_shadow(linestring)

    def merge_shaded_areas(self):
        """Merge shaded areas of all PV segments"""
        for seg in self.list_segments:
            seg._shaded_collection.merge_surfaces()

    def cut_at_point(self, point):
        """Cut Side at point if the side contains it.

        Parameters
        ----------
        point : :py:class:`shapely.geometry.Point`
            Point where to cut side geometry, if the latter contains the
            former
        """
        if contains(self, point):
            for segment in self.list_segments:
                # Nothing will happen to the segments that do not contain
                # the point
                segment.cut_at_point(point)

    def get_param_weighted(self, param):
        """Get the parameter from the side's surfaces, after weighting
        by surface length.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        float
            Weighted parameter value
        """
        value = self.get_param_ww(param) / self.length
        return value

    def get_param_ww(self, param):
        """Get the parameter from the side's surfaces with weight, i.e.
        after multiplying by the surface lengths.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        float
            Parameter value multiplied by weights

        Raises
        ------
        KeyError
            if parameter name not in a surface parameters
        """
        value = 0
        for seg in self.list_segments:
            value += seg.get_param_ww(param)
        return value

    def update_params(self, new_dict):
        """Update surface parameters in the Side.

        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surfaces
        """
        for seg in self.list_segments:
            seg.update_params(new_dict)


class BasePVArray(object):
    """Base class for PV arrays in pvfactors. Will provide basic
    capabilities."""

    registry_cols = ['geom', 'line_type', 'pvrow_index', 'side',
                     'pvsegment_index', 'shaded', 'surface_index']

    def __init__(self, axis_azimuth=None):
        """Initialize Base of PV array.

        Parameters
        ----------
        axis_azimuth : float, optional
            Azimuth angle of rotation axis [deg] (Default = None)
        """
        # All PV arrays should have a fixed axis azimuth in pvfactors
        self.axis_azimuth = axis_azimuth

        # The are required attributes of any PV array
        self.ts_pvrows = None
        self.ts_ground = None

    @property
    def n_ts_surfaces(self):
        """Number of timeseries surfaces in the PV array."""
        n_ts_surfaces = 0
        n_ts_surfaces += self.ts_ground.n_ts_surfaces
        for ts_pvrow in self.ts_pvrows:
            n_ts_surfaces += ts_pvrow.n_ts_surfaces
        return n_ts_surfaces

    @property
    def all_ts_surfaces(self):
        """List of all timeseries surfaces in PV array"""
        all_ts_surfaces = []
        all_ts_surfaces += self.ts_ground.all_ts_surfaces
        for ts_pvrow in self.ts_pvrows:
            all_ts_surfaces += ts_pvrow.all_ts_surfaces
        return all_ts_surfaces

    @property
    def ts_surface_indices(self):
        """List of indices of all the timeseries surfaces"""
        return [ts_surf.index for ts_surf in self.all_ts_surfaces]

    def plot_at_idx(self, idx, ax, merge_if_flag_overlap=True,
                    with_cut_points=True, x_min_max=None,
                    with_surface_index=False):
        """Plot all the PV rows and the ground in the PV array at a desired
        step index. This can be called before transforming the array, and
        after fitting it.

        Parameters
        ----------
        idx : int
            Selected timestep index for plotting the PV array
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting the PV array geometries
        merge_if_flag_overlap : bool, optional
            Decide whether to merge all shadows if they overlap
            (Default = True)
        with_cut_points :  bool, optional
            Decide whether to include the saved cut points in the created
            PV ground geometry (Default = True)
        x_min_max : tuple, optional
            List of minimum and maximum x coordinates for the flat ground
            surface [m] (Default = None)
        with_surface_index : bool, optional
            Plot the surfaces with their index values (Default = False)
        """
        # Plot pv array structures
        self.ts_ground.plot_at_idx(
            idx, ax, color_shaded=COLOR_DIC['ground_shaded'],
            color_illum=COLOR_DIC['ground_illum'],
            merge_if_flag_overlap=merge_if_flag_overlap,
            with_cut_points=with_cut_points, x_min_max=x_min_max,
            with_surface_index=with_surface_index)

        for ts_pvrow in self.ts_pvrows:
            ts_pvrow.plot_at_idx(
                idx, ax, color_shaded=COLOR_DIC['pvrow_shaded'],
                color_illum=COLOR_DIC['pvrow_illum'],
                with_surface_index=with_surface_index)

        # Plot formatting
        ax.axis('equal')
        if self.distance is not None:
            n_pvrows = self.n_pvrows
            ax.set_xlim(- 0.5 * self.distance,
                        (n_pvrows - 0.5) * self.distance)
        if self.height is not None:
            ax.set_ylim(- self.height, 2 * self.height)
        ax.set_xlabel("x [m]", fontsize=PLOT_FONTSIZE)
        ax.set_ylabel("y [m]", fontsize=PLOT_FONTSIZE)

    def fit(self, *args, **kwargs):
        """Not implemented."""
        raise NotImplementedError

    def update_params(self, new_dict):
        """Update timeseries surface parameters in the collection.
        Parameters
        ----------
        new_dict : dict
            Parameters to add or update for the surfaces
        """
        self.ts_ground.update_params(new_dict)
        for ts_pvrow in self.ts_pvrows:
            ts_pvrow.update_params(new_dict)

    def _index_all_ts_surfaces(self):
        """Add unique indices to all surfaces in the PV array."""
        for idx, ts_surface in enumerate(self.all_ts_surfaces):
            ts_surface.index = idx



"""Module containing irradiance models used with pv array geometries"""

from pvlib.tools import cosd
from pvlib.irradiance import aoi as aoi_function
from pvlib.irradiance import get_total_irradiance
import numpy as np
from pvfactors.irradiance.utils import \
    perez_diffuse_luminance, calculate_horizon_band_shading, \
    calculate_circumsolar_shading
from pvfactors.irradiance.base import BaseModel
from pvfactors.config import \
    DEFAULT_HORIZON_BAND_ANGLE, SKY_REFLECTIVITY_DUMMY, \
    DEFAULT_CIRCUMSOLAR_ANGLE


class IsotropicOrdered(BaseModel):
    """Diffuse isotropic sky model for
    :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`. It will
    calculate the appropriate values for an isotropic sky dome and apply
    it to the PV array."""

    params = ['rho', 'inv_rho', 'direct', 'isotropic', 'reflection']
    cats = ['ground_illum', 'ground_shaded', 'front_illum_pvrow',
            'back_illum_pvrow', 'front_shaded_pvrow', 'back_shaded_pvrow']
    irradiance_comp = ['direct']
    irradiance_comp_absorbed = ['direct_absorbed']

    def __init__(self, rho_front=0.01, rho_back=0.03, module_transparency=0.,
                 module_spacing_ratio=0., faoi_fn_front=None,
                 faoi_fn_back=None):
        """Initialize irradiance model values that will be saved later on.

        Parameters
        ----------
        rho_front : float, optional
            Reflectivity of the front side of the PV rows (default = 0.01)
        rho_back : float, optional
            Reflectivity of the back side of the PV rows (default = 0.03)
        module_transparency : float, optional
            Module transparency (from 0 to 1), which will let some direct light
            pass through the PV modules in the PV rows and reach the shaded
            ground (Default = 0., fully opaque)
        module_spacing_ratio : float, optional
            Module spacing ratio (from 0 to 1), which is the ratio of the area
            covered by the space between PV modules over the total area of the
            PV rows, and which determines how much direct light will reach the
            shaded ground through the PV rows
            (Default = 0., no spacing at all)
        faoi_fn_front : function or object, optional
            Function (or object containing ``faoi`` method)
            which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the front side of PV rows (default = None)
        faoi_fn_back : function or object, optional
            Function (or object containing ``faoi`` method)
            which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the back side of PV rows (default = None)
        """
        self.direct = dict.fromkeys(self.cats)
        self.total_perez = dict.fromkeys(self.cats)
        self.faoi_front = dict.fromkeys(self.irradiance_comp)
        self.faoi_back = dict.fromkeys(self.irradiance_comp)
        self.faoi_ground = None
        self.module_transparency = module_transparency
        self.module_spacing_ratio = module_spacing_ratio
        self.rho_front = rho_front
        self.rho_back = rho_back
        # Treatment of faoi functions
        faoi_fn_front = (faoi_fn_front.faoi if hasattr(faoi_fn_front, 'faoi')
                         else faoi_fn_front)
        faoi_fn_back = (faoi_fn_back.faoi if hasattr(faoi_fn_back, 'faoi')
                        else faoi_fn_back)
        self.faoi_fn_front = faoi_fn_front
        self.faoi_fn_back = faoi_fn_back
        # The following will be updated at fitting time
        self.albedo = None
        self.GHI = None
        self.DHI = None
        self.isotropic_luminance = None

    def fit(self, timestamps, DNI, DHI, solar_zenith, solar_azimuth,
            surface_tilt, surface_azimuth, albedo,
            ghi=None):
        """Use vectorization to calculate values used for the isotropic
        irradiance model.

        Parameters
        ----------
        timestamps : array-like
            List of timestamps of the simulation.
        DNI : array-like
            Direct normal irradiance values [W/m2]
        DHI : array-like
            Diffuse horizontal irradiance values [W/m2]
        solar_zenith : array-like
            Solar zenith angles [deg]
        solar_azimuth : array-like
            Solar azimuth angles [deg]
        surface_tilt : array-like
            Surface tilt angles, from 0 to 180 [deg]
        surface_azimuth : array-like
            Surface azimuth angles [deg]
        albedo : array-like
            Albedo values (or ground reflectivity)
        ghi : array-like, optional
            Global horizontal irradiance [W/m2], if not provided, will be
            calculated from DNI and DHI (Default = None)
        """
        # Make sure getting array-like values
        if np.isscalar(DNI):
            timestamps = [timestamps]
            DNI = np.array([DNI])
            DHI = np.array([DHI])
            solar_zenith = np.array([solar_zenith])
            solar_azimuth = np.array([solar_azimuth])
            surface_tilt = np.array([surface_tilt])
            surface_azimuth = np.array([surface_azimuth])
            ghi = None if ghi is None else np.array([ghi])
        # Length of arrays
        n = len(DNI)
        # Make sure that albedo is a vector
        albedo = albedo * np.ones(n) if np.isscalar(albedo) else albedo

        # Save and calculate total POA values from Perez model
        ghi = DNI * cosd(solar_zenith) + DHI if ghi is None else ghi
        self.GHI = ghi
        self.DHI = DHI
        self.n_steps = n
        perez_front_pvrow = get_total_irradiance(
            surface_tilt, surface_azimuth, solar_zenith, solar_azimuth,
            DNI, ghi, DHI, albedo=albedo)

        # Save diffuse light
        self.isotropic_luminance = DHI
        self.albedo = albedo

        # DNI seen by ground illuminated surfaces
        self.direct['ground_illum'] = DNI * cosd(solar_zenith)
        self.direct['ground_shaded'] = (
            # Direct light through PV modules spacing
            self.direct['ground_illum'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.direct['ground_illum'] * (1. - self.module_spacing_ratio)
            * self.module_transparency)

        # Calculate AOI on front pvrow using pvlib implementation
        aoi_front_pvrow = aoi_function(
            surface_tilt, surface_azimuth, solar_zenith, solar_azimuth)
        aoi_back_pvrow = 180. - aoi_front_pvrow

        # calculate faoi modifiers
        self._calculate_faoi_modifiers(aoi_front_pvrow, aoi_back_pvrow,
                                       surface_tilt, albedo)

        # DNI seen by pvrow illuminated surfaces
        front_is_illum = aoi_front_pvrow <= 90
        # direct
        self.direct['front_illum_pvrow'] = np.where(
            front_is_illum, DNI * cosd(aoi_front_pvrow), 0.)
        self.direct['front_shaded_pvrow'] = (
            # Direct light through PV modules spacing
            self.direct['front_illum_pvrow'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.direct['front_illum_pvrow']
            * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        self.direct['back_illum_pvrow'] = np.where(
            ~front_is_illum, DNI * cosd(aoi_back_pvrow), 0.)
        self.direct['back_shaded_pvrow'] = (
            # Direct light through PV modules spacing
            self.direct['back_illum_pvrow'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.direct['back_illum_pvrow']
            * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        # perez
        self.total_perez['front_illum_pvrow'] = perez_front_pvrow['poa_global']
        self.total_perez['front_shaded_pvrow'] = (
            self.total_perez['front_illum_pvrow']
            - self.direct['front_illum_pvrow'])
        self.total_perez['ground_shaded'] = (self.DHI
                                             + self.direct['ground_shaded'])

    def transform(self, pvarray):
        """Apply calculated irradiance values to PV array timeseries
        geometries: assign values as parameters to timeseries surfaces.

        Parameters
        ----------
        pvarray : PV array object
            PV array on which the calculated irradiance values will be applied
        """

        # Prepare variables
        n_steps = self.n_steps
        rho_front = self.rho_front * np.ones(n_steps)
        inv_rho_front = 1. / rho_front
        rho_back = self.rho_back * np.ones(n_steps)
        inv_rho_back = 1. / rho_back

        # Transform timeseries ground
        pvarray.ts_ground.update_illum_params(
            {'direct': self.direct['ground_illum'],
             'rho': self.albedo,
             'inv_rho': 1. / self.albedo,
             'total_perez': self.gnd_illum,
             'direct_absorbed': self.faoi_ground * self.direct['ground_illum']
             })
        pvarray.ts_ground.update_shaded_params(
            {'direct': self.direct['ground_shaded'],
             'rho': self.albedo,
             'inv_rho': 1. / self.albedo,
             'total_perez': self.gnd_shaded,
             'direct_absorbed': self.faoi_ground * self.direct['ground_shaded']
             })

        for ts_pvrow in pvarray.ts_pvrows:
            # Front
            for ts_seg in ts_pvrow.front.list_segments:
                ts_seg.illum.update_params(
                    {'direct': self.direct['front_illum_pvrow'],
                     'rho': rho_front,
                     'inv_rho': inv_rho_front,
                     'total_perez': self.pvrow_illum,
                     'direct_absorbed': self.faoi_front['direct']
                     * self.direct['front_illum_pvrow']})
                ts_seg.shaded.update_params(
                    {'direct': self.direct['front_shaded_pvrow'],
                     'rho': rho_front,
                     'inv_rho': inv_rho_front,
                     'total_perez': self.pvrow_shaded,
                     'direct_absorbed': self.faoi_front['direct']
                     * self.direct['front_shaded_pvrow']})
            # Back
            for ts_seg in ts_pvrow.back.list_segments:
                ts_seg.illum.update_params(
                    {'direct': self.direct['back_illum_pvrow'],
                     'rho': rho_back,
                     'inv_rho': inv_rho_back,
                     'total_perez': np.zeros(n_steps),
                     'direct_absorbed': self.faoi_back['direct']
                     * self.direct['back_illum_pvrow']})
                ts_seg.shaded.update_params(
                    {'direct': self.direct['back_shaded_pvrow'],
                     'rho': rho_back,
                     'inv_rho': inv_rho_back,
                     'total_perez': np.zeros(n_steps),
                     'direct_absorbed': self.faoi_back['direct']
                     * self.direct['back_shaded_pvrow']})

    def get_full_modeling_vectors(self, pvarray, idx):
        """Get the modeling vectors used in matrix calculations of mathematical
        model.

        Parameters
        ----------
        pvarray : PV array object
            PV array on which the calculated irradiance values will be applied
        idx : int, optional
            Index of the irradiance values to apply to the PV array (in the
            whole timeseries values)

        Returns
        -------
        irradiance_vec : numpy array
            List of summed up non-reflective irradiance values for all surfaces
            and sky
        rho_vec : numpy array
            List of reflectivity values for all surfaces and sky
        invrho_vec : numpy array
            List of inverse reflectivity for all surfaces and sky
        total_perez_vec : numpy array
            List of total perez transposed irradiance values for all surfaces
            and sky
        """
        # Sum up the necessary parameters to form the irradiance vector
        irradiance_vec, rho_vec, inv_rho_vec, total_perez_vec = \
            self.get_modeling_vectors(pvarray)
        # Add sky values
        irradiance_vec.append(self.isotropic_luminance[idx])
        total_perez_vec.append(self.isotropic_luminance[idx])
        rho_vec.append(SKY_REFLECTIVITY_DUMMY)
        inv_rho_vec.append(SKY_REFLECTIVITY_DUMMY)

        return np.array(irradiance_vec), np.array(rho_vec), \
            np.array(inv_rho_vec), np.array(total_perez_vec)

    def get_full_ts_modeling_vectors(self, pvarray):
        """Get the modeling vectors used in matrix calculations of
        the mathematical model, including the sky values.

        Parameters
        ----------
        pvarray : PV array object
            PV array on which the calculated irradiance values will be applied

        Returns
        -------
        irradiance_mat : np.ndarray
            List of summed up non-reflective irradiance values for all surfaces
            and sky. Dimension = [n_surfaces + 1, n_timesteps]
        rho_mat : np.ndarray
            List of reflectivity values for all surfaces and sky.
            Dimension = [n_surfaces + 1, n_timesteps]
        invrho_mat : np.ndarray
            List of inverse reflectivity for all surfaces and sky.
            Dimension = [n_surfaces + 1, n_timesteps]
        total_perez_mat : np.ndarray
            List of total perez transposed irradiance values for all surfaces
            and sky. Dimension = [n_surfaces + 1, n_timesteps]
        """
        # Sum up the necessary parameters to form the irradiance vector
        irradiance_mat, rho_mat, inv_rho_mat, total_perez_mat = \
            self.get_ts_modeling_vectors(pvarray)
        # Add sky values
        irradiance_mat.append(self.isotropic_luminance)
        total_perez_mat.append(self.isotropic_luminance)
        rho_mat.append(SKY_REFLECTIVITY_DUMMY * np.ones(pvarray.n_states))
        inv_rho_mat.append(SKY_REFLECTIVITY_DUMMY * np.ones(pvarray.n_states))

        return np.array(irradiance_mat), np.array(rho_mat), \
            np.array(inv_rho_mat), np.array(total_perez_mat)

    @property
    def gnd_shaded(self):
        """Total timeseries irradiance incident on ground shaded areas"""
        return self.total_perez['ground_shaded']

    @property
    def gnd_illum(self):
        """Total timeseries irradiance incident on ground illuminated areas"""
        return self.GHI

    @property
    def pvrow_shaded(self):
        """Total timeseries irradiance incident on PV row's front illuminated
        areas and calculated by Perez transposition"""
        return self.total_perez['front_shaded_pvrow']

    @property
    def pvrow_illum(self):
        """Total timeseries irradiance incident on PV row's front shaded
        areas and calculated by Perez transposition"""
        return self.total_perez['front_illum_pvrow']

    @property
    def sky_luminance(self):
        """Total timeseries isotropic luminance of sky"""
        return self.isotropic_luminance

    def _calculate_faoi_modifiers(self, aoi_front_pvrow, aoi_back_pvrow,
                                  surface_tilt, albedo):
        """Calculate fAOI modifier values for all surface types: front PV row,
        back PV row, and ground.

        Parameters
        ----------
        aoi_front_pvrow : np.ndarray
            Direct light angle of incidence values for PV row front side [deg]
        aoi_back_pvrow : np.ndarray
            Direct light angle of incidence values for PV row back side [deg]
        surface_tilt : np.ndarray
            Surface tilt of PV row surfaces, measured from horizontal,
            ranging from 0 to 180 [deg]
        albedo : np.ndarray
            Ground albedo values
        """
        # Need to update aoi measurement: input measured from normal,
        # but functions require to measure it from surface horizontal
        aoi_front = np.abs(90. - aoi_front_pvrow)
        aoi_back = np.abs(90. - aoi_back_pvrow)
        # --- front
        self.faoi_front['direct'] = (
            self.faoi_fn_front(aoi_front)
            if self.faoi_fn_front is not None
            else (1. - self.rho_front) * np.ones_like(surface_tilt))

        # --- back
        self.faoi_back['direct'] = (
            self.faoi_fn_back(aoi_back)
            if self.faoi_fn_back is not None
            else (1. - self.rho_back) * np.ones_like(surface_tilt))

        # --- ground
        self.faoi_ground = (1. - albedo)


class HybridPerezOrdered(BaseModel):
    """Model is based off Perez diffuse light model, and
    applied to pvfactors :py:class:`~pvfactors.geometry.pvarray.OrderedPVArray`
    objects.
    The model applies direct, circumsolar, and horizon irradiance to the PV
    array surfaces.
    """

    params = ['rho', 'inv_rho', 'direct', 'isotropic', 'circumsolar',
              'horizon', 'reflection']
    cats = ['ground_illum', 'ground_shaded', 'front_illum_pvrow',
            'back_illum_pvrow', 'front_shaded_pvrow', 'back_shaded_pvrow']
    irradiance_comp = ['direct', 'circumsolar', 'horizon']
    irradiance_comp_absorbed = ['direct_absorbed', 'circumsolar_absorbed',
                                'horizon_absorbed']

    def __init__(self, horizon_band_angle=DEFAULT_HORIZON_BAND_ANGLE,
                 circumsolar_angle=DEFAULT_CIRCUMSOLAR_ANGLE,
                 circumsolar_model='uniform_disk', rho_front=0.01,
                 rho_back=0.03, module_transparency=0.,
                 module_spacing_ratio=0., faoi_fn_front=None,
                 faoi_fn_back=None):
        """Initialize irradiance model values that will be saved later on.

        Parameters
        ----------
        horizon_band_angle : float, optional
            Width of the horizon band in [deg] (Default =
            DEFAULT_HORIZON_BAND_ANGLE)
        circumsolar_angle : float, optional
            Diameter of the circumsolar area in [deg] (Default =
            DEFAULT_CIRCUMSOLAR_ANGLE)
        circumsolar_model : str
            Circumsolar shading model to use (Default = 'uniform_disk')
        rho_front : float, optional
            Reflectivity of the front side of the PV rows (default = 0.01)
        rho_back : float, optional
            Reflectivity of the back side of the PV rows (default = 0.03)
        module_transparency : float, optional
            Module transparency (from 0 to 1), which will let some direct light
            pass through the PV modules in the PV rows and reach the shaded
            ground (Default = 0., fully opaque)
        module_spacing_ratio : float, optional
            Module spacing ratio (from 0 to 1), which is the ratio of the area
            covered by the space between PV modules over the total area of the
            PV rows, and which determines how much direct light will reach the
            shaded ground through the PV rows
            (Default = 0., no spacing at all)
        faoi_fn_front : function or object, optional
            Function (or object containing ``faoi`` method)
            which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the front side of PV rows (default = None)
        faoi_fn_back : function or object, optional
            Function (or object containing ``faoi`` method)
            which takes a list (or numpy array) of incidence angles
            measured from the surface horizontal
            (with values from 0 to 180 deg) and returns the fAOI values for
            the back side of PV rows (default = None)
        """
        self.direct = dict.fromkeys(self.cats)
        self.circumsolar = dict.fromkeys(self.cats)
        self.horizon = dict.fromkeys(self.cats)
        self.total_perez = dict.fromkeys(self.cats)
        self.faoi_front = dict.fromkeys(self.irradiance_comp)
        self.faoi_back = dict.fromkeys(self.irradiance_comp)
        self.faoi_ground = None
        self.horizon_band_angle = horizon_band_angle
        self.circumsolar_angle = circumsolar_angle
        self.circumsolar_model = circumsolar_model
        self.rho_front = rho_front
        self.rho_back = rho_back
        self.module_transparency = module_transparency
        self.module_spacing_ratio = module_spacing_ratio
        # Treatment of faoi functions
        faoi_fn_front = (faoi_fn_front.faoi if hasattr(faoi_fn_front, 'faoi')
                         else faoi_fn_front)
        faoi_fn_back = (faoi_fn_back.faoi if hasattr(faoi_fn_back, 'faoi')
                        else faoi_fn_back)
        self.faoi_fn_front = faoi_fn_front
        self.faoi_fn_back = faoi_fn_back
        # Attributes that will be updated at fitting time
        self.albedo = None
        self.GHI = None
        self.DNI = None
        self.n_steps = None
        self.isotropic_luminance = None

    def fit(self, timestamps, DNI, DHI, solar_zenith, solar_azimuth,
            surface_tilt, surface_azimuth, albedo,
            ghi=None):
        """Use vectorization to calculate values used for the hybrid Perez
        irradiance model.

        Parameters
        ----------
        timestamps : array-like
            List of timestamps of the simulation.
        DNI : array-like
            Direct normal irradiance values [W/m2]
        DHI : array-like
            Diffuse horizontal irradiance values [W/m2]
        solar_zenith : array-like
            Solar zenith angles [deg]
        solar_azimuth : array-like
            Solar azimuth angles [deg]
        surface_tilt : array-like
            Surface tilt angles, from 0 to 180 [deg]
        surface_azimuth : array-like
            Surface azimuth angles [deg]
        albedo : array-like
            Albedo values (or ground reflectivity)
        ghi : array-like, optional
            Global horizontal irradiance [W/m2], if not provided, will be
            calculated from DNI and DHI (Default = None)
        """
        # Make sure getting array-like values
        if np.isscalar(DNI):
            timestamps = [timestamps]
            DNI = np.array([DNI])
            DHI = np.array([DHI])
            solar_zenith = np.array([solar_zenith])
            solar_azimuth = np.array([solar_azimuth])
            surface_tilt = np.array([surface_tilt])
            surface_azimuth = np.array([surface_azimuth])
            ghi = None if ghi is None else np.array([ghi])
        # Length of arrays
        n = len(DNI)
        # Make sure that albedo is a vector
        albedo = albedo * np.ones(n) if np.isscalar(albedo) else albedo

        # Calculate terms from Perez model
        luminance_isotropic, luminance_circumsolar, poa_horizon, \
            poa_circumsolar_front, poa_circumsolar_back, \
            aoi_front_pvrow, aoi_back_pvrow = \
            self._calculate_luminance_poa_components(
                timestamps, DNI, DHI, solar_zenith, solar_azimuth,
                surface_tilt, surface_azimuth)

        # calculate faoi modifiers
        self._calculate_faoi_modifiers(aoi_front_pvrow, aoi_back_pvrow,
                                       surface_tilt, albedo)

        # Save and calculate total POA values from Perez model
        ghi = DNI * cosd(solar_zenith) + DHI if ghi is None else ghi
        self.GHI = ghi
        self.DHI = DHI
        self.n_steps = n
        perez_front_pvrow = get_total_irradiance(
            surface_tilt, surface_azimuth, solar_zenith, solar_azimuth,
            DNI, ghi, DHI, albedo=albedo)
        total_perez_front_pvrow = perez_front_pvrow['poa_global']

        # Save isotropic luminance
        self.isotropic_luminance = luminance_isotropic
        self.albedo = albedo

        # Ground surfaces
        self.direct['ground_illum'] = DNI * cosd(solar_zenith)
        self.direct['ground_shaded'] = (
            # Direct light through PV modules spacing
            self.direct['ground_illum'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.direct['ground_illum'] * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        self.circumsolar['ground_illum'] = luminance_circumsolar
        self.circumsolar['ground_shaded'] = (
            # Circumsolar light through PV modules spacing
            self.circumsolar['ground_illum'] * self.module_spacing_ratio
            # Circumsolar light through PV modules, by transparency
            + self.circumsolar['ground_illum']
            * (1. - self.module_spacing_ratio) * self.module_transparency)
        self.horizon['ground'] = np.zeros(n)

        # PV row surfaces
        front_is_illum = aoi_front_pvrow <= 90
        # direct
        self.direct['front_illum_pvrow'] = np.where(
            front_is_illum, DNI * cosd(aoi_front_pvrow), 0.)
        self.direct['front_shaded_pvrow'] = (
            # Direct light through PV modules spacing
            self.direct['front_illum_pvrow'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.direct['front_illum_pvrow']
            * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        self.direct['back_illum_pvrow'] = np.where(
            ~front_is_illum, DNI * cosd(aoi_back_pvrow), 0.)
        self.direct['back_shaded_pvrow'] = (
            # Direct light through PV modules spacing
            self.direct['back_illum_pvrow'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.direct['back_illum_pvrow']
            * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        # circumsolar
        self.circumsolar['front_illum_pvrow'] = np.where(
            front_is_illum, poa_circumsolar_front, 0.)
        self.circumsolar['front_shaded_pvrow'] = (
            # Direct light through PV modules spacing
            self.circumsolar['front_illum_pvrow'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.circumsolar['front_illum_pvrow']
            * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        self.circumsolar['back_illum_pvrow'] = np.where(
            ~front_is_illum, poa_circumsolar_back, 0.)
        self.circumsolar['back_shaded_pvrow'] = (
            # Direct light through PV modules spacing
            self.circumsolar['back_illum_pvrow'] * self.module_spacing_ratio
            # Direct light through PV modules, by transparency
            + self.circumsolar['back_illum_pvrow']
            * (1. - self.module_spacing_ratio)
            * self.module_transparency)
        # horizon
        self.horizon['front_pvrow'] = poa_horizon
        self.horizon['back_pvrow'] = poa_horizon
        # perez
        self.total_perez['front_illum_pvrow'] = total_perez_front_pvrow
        self.total_perez['front_shaded_pvrow'] = (
            total_perez_front_pvrow
            - self.direct['front_illum_pvrow']
            - self.circumsolar['front_illum_pvrow']
            + self.direct['front_shaded_pvrow']
            + self.circumsolar['front_shaded_pvrow']
        )
        self.total_perez['ground_shaded'] = (
            DHI
            - self.circumsolar['ground_illum']
            + self.circumsolar['ground_shaded']
            + self.direct['ground_shaded'])
        self.total_perez['ground_illum'] = ghi
        self.total_perez['sky'] = luminance_isotropic

    def transform(self, pvarray):
        """Apply calculated irradiance values to PV array timeseries
        geometries: assign values as parameters to timeseries surfaces.

        Parameters
        ----------
        pvarray : PV array object
            PV array on which the calculated irradiance values will be applied
        """

        # Prepare variables
        n_steps = self.n_steps
        ts_pvrows = pvarray.ts_pvrows
        tilted_to_left = pvarray.rotation_vec > 0
        rho_front = self.rho_front * np.ones(n_steps)
        inv_rho_front = 1. / rho_front
        rho_back = self.rho_back * np.ones(n_steps)
        inv_rho_back = 1. / rho_back

        # Transform timeseries ground
        pvarray.ts_ground.update_illum_params({
            'direct': self.direct['ground_illum'],
            'circumsolar': self.circumsolar['ground_illum'],
            'horizon': np.zeros(n_steps),
            'rho': self.albedo,
            'inv_rho': 1. / self.albedo,
            'total_perez': self.gnd_illum,
            'direct_absorbed': self.faoi_ground * self.direct['ground_illum'],
            'circumsolar_absorbed': self.faoi_ground
            * self.circumsolar['ground_illum'],
            'horizon_absorbed': np.zeros(n_steps)})
        pvarray.ts_ground.update_shaded_params({
            'direct': self.direct['ground_shaded'],
            'circumsolar': self.circumsolar['ground_shaded'],
            'horizon': np.zeros(n_steps),
            'rho': self.albedo,
            'inv_rho': 1. / self.albedo,
            'total_perez': self.gnd_shaded,
            'direct_absorbed': self.faoi_ground * self.direct['ground_shaded'],
            'circumsolar_absorbed': self.faoi_ground
            * self.circumsolar['ground_shaded'],
            'horizon_absorbed': np.zeros(n_steps)})

        # Transform timeseries PV rows
        for idx_pvrow, ts_pvrow in enumerate(ts_pvrows):
            # Front
            for ts_seg in ts_pvrow.front.list_segments:
                ts_seg.illum.update_params({
                    'direct': self.direct['front_illum_pvrow'],
                    'circumsolar': self.circumsolar['front_illum_pvrow'],
                    'horizon': self.horizon['front_pvrow'],
                    'rho': rho_front,
                    'inv_rho': inv_rho_front,
                    'total_perez': self.pvrow_illum,
                    'direct_absorbed': self.faoi_front['direct']
                    * self.direct['front_illum_pvrow'],
                    'circumsolar_absorbed': self.faoi_front['circumsolar']
                    * self.circumsolar['front_illum_pvrow'],
                    'horizon_absorbed': self.faoi_front['horizon']
                    * self.horizon['front_pvrow']})
                ts_seg.shaded.update_params({
                    'direct': self.direct['front_shaded_pvrow'],
                    'circumsolar': self.circumsolar['front_shaded_pvrow'],
                    'horizon': self.horizon['front_pvrow'],
                    'rho': rho_front,
                    'inv_rho': inv_rho_front,
                    'total_perez': self.pvrow_shaded,
                    'direct_absorbed': self.faoi_front['direct']
                    * self.direct['front_shaded_pvrow'],
                    'circumsolar_absorbed': self.faoi_front['circumsolar']
                    * self.circumsolar['front_shaded_pvrow'],
                    'horizon_absorbed': self.faoi_front['horizon']
                    * self.horizon['front_pvrow']})
            # Back: apply back surface horizon shading
            for ts_seg in ts_pvrow.back.list_segments:
                # Illum
                # In ordered pv arrays, there should only be 1 surface -> 0
                centroid_illum = ts_seg.illum.list_ts_surfaces[0].centroid
                hor_shd_pct_illum = self._calculate_horizon_shading_pct_ts(
                    ts_pvrows, centroid_illum, idx_pvrow, tilted_to_left,
                    is_back_side=True)
                ts_seg.illum.update_params({
                    'direct': self.direct['back_illum_pvrow'],
                    'circumsolar': self.circumsolar['back_illum_pvrow'],
                    'horizon': self.horizon['back_pvrow'] *
                    (1. - hor_shd_pct_illum / 100.),
                    'horizon_unshaded': self.horizon['back_pvrow'],
                    'horizon_shd_pct': hor_shd_pct_illum,
                    'rho': rho_back,
                    'inv_rho': inv_rho_back,
                    'total_perez': np.zeros(n_steps),
                    'direct_absorbed': self.faoi_back['direct']
                    * self.direct['back_illum_pvrow'],
                    'circumsolar_absorbed': self.faoi_back['circumsolar']
                    * self.circumsolar['back_illum_pvrow'],
                    'horizon_absorbed': self.faoi_back['horizon']
                    * self.horizon['back_pvrow']
                    * (1. - hor_shd_pct_illum / 100.)})
                # Back
                # In ordered pv arrays, there should only be 1 surface -> 0
                centroid_shaded = ts_seg.shaded.list_ts_surfaces[0].centroid
                hor_shd_pct_shaded = self._calculate_horizon_shading_pct_ts(
                    ts_pvrows, centroid_shaded, idx_pvrow, tilted_to_left,
                    is_back_side=True)
                ts_seg.shaded.update_params({
                    'direct': self.direct['back_shaded_pvrow'],
                    'circumsolar': self.circumsolar['back_shaded_pvrow'],
                    'horizon': self.horizon['back_pvrow'] *
                    (1. - hor_shd_pct_shaded / 100.),
                    'horizon_unshaded': self.horizon['back_pvrow'],
                    'horizon_shd_pct': hor_shd_pct_shaded,
                    'rho': rho_back,
                    'inv_rho': inv_rho_back,
                    'total_perez': np.zeros(n_steps),
                    'direct_absorbed': self.faoi_back['direct']
                    * self.direct['back_shaded_pvrow'],
                    'circumsolar_absorbed': self.faoi_back['circumsolar']
                    * self.circumsolar['back_shaded_pvrow'],
                    'horizon_absorbed': self.faoi_back['horizon']
                    * self.horizon['back_pvrow']
                    * (1. - hor_shd_pct_shaded / 100.)})

    def get_full_modeling_vectors(self, pvarray, idx):
        """Get the modeling vectors used in matrix calculations of mathematical
        model.

        Parameters
        ----------
        pvarray : PV array object
            PV array on which the calculated irradiance values will be applied
        idx : int, optional
            Index of the irradiance values to apply to the PV array (in the
            whole timeseries values)

        Returns
        -------
        irradiance_vec : numpy array
            List of summed up non-reflective irradiance values for all surfaces
            and sky
        rho_vec : numpy array
            List of reflectivity values for all surfaces and sky
        invrho_vec : numpy array
            List of inverse reflectivity for all surfaces and sky
        total_perez_vec : numpy array
            List of total perez transposed irradiance values for all surfaces
            and sky
        """

        # Sum up the necessary parameters to form the irradiance vector
        irradiance_vec, rho_vec, inv_rho_vec, total_perez_vec = self.get_modeling_vectors(pvarray)
        # Add sky values
        irradiance_vec.append(self.isotropic_luminance[idx])
        total_perez_vec.append(self.isotropic_luminance[idx])
        rho_vec.append(SKY_REFLECTIVITY_DUMMY)
        inv_rho_vec.append(SKY_REFLECTIVITY_DUMMY)

        return np.array(irradiance_vec), np.array(rho_vec), \
            np.array(inv_rho_vec), np.array(total_perez_vec)

    def get_full_ts_modeling_vectors(self, pvarray):
        """Get the modeling vectors used in matrix calculations of
        the mathematical model, including the sky values.

        Parameters
        ----------
        pvarray : PV array object
            PV array on which the calculated irradiance values will be applied

        Returns
        -------
        irradiance_mat : np.ndarray
            List of summed up non-reflective irradiance values for all surfaces
            and sky. Dimension = [n_surfaces + 1, n_timesteps]
        rho_mat : np.ndarray
            List of reflectivity values for all surfaces and sky.
            Dimension = [n_surfaces + 1, n_timesteps]
        invrho_mat : np.ndarray
            List of inverse reflectivity for all surfaces and sky.
            Dimension = [n_surfaces + 1, n_timesteps]
        total_perez_mat : np.ndarray
            List of total perez transposed irradiance values for all surfaces
            and sky. Dimension = [n_surfaces + 1, n_timesteps]
        """
        # Sum up the necessary parameters to form the irradiance vector
        irradiance_mat, rho_mat, inv_rho_mat, total_perez_mat = self.get_ts_modeling_vectors(pvarray)
        # Add sky values
        irradiance_mat.append(self.isotropic_luminance)
        total_perez_mat.append(self.isotropic_luminance)
        rho_mat.append(SKY_REFLECTIVITY_DUMMY * np.ones(pvarray.n_states))
        inv_rho_mat.append(SKY_REFLECTIVITY_DUMMY * np.ones(pvarray.n_states))

        return np.array(irradiance_mat), np.array(rho_mat), \
            np.array(inv_rho_mat), np.array(total_perez_mat)

    @property
    def gnd_shaded(self):
        """Total timeseries irradiance incident on ground shaded areas"""
        return self.total_perez['ground_shaded']

    @property
    def gnd_illum(self):
        """Total timeseries irradiance incident on ground illuminated areas"""
        return self.total_perez['ground_illum']

    @property
    def pvrow_shaded(self):
        """Total timeseries irradiance incident on PV row's front illuminated
        areas and calculated by Perez transposition"""
        return self.total_perez['front_shaded_pvrow']

    @property
    def pvrow_illum(self):
        """Total timeseries irradiance incident on PV row's front shaded
        areas and calculated by Perez transposition"""
        return self.total_perez['front_illum_pvrow']

    @property
    def sky_luminance(self):
        """Total timeseries isotropic luminance of sky"""
        return self.isotropic_luminance

    def _calculate_horizon_shading_pct_ts(self, ts_pvrows, ts_point_coords,
                                          pvrow_idx, tilted_to_left,
                                          is_back_side=True):
        """Calculate horizon band shading percentage on surfaces of the ordered
        PV array, in a vectorized way.

        Parameters
        ----------
        ts_pvrows : list of :py:class:`~pvfactors.geometry.timeseries.TsPVRow`
            List of timeseries PV rows in the PV array
        ts_point_coords : :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of point that suffers horizon band shading
        pvrow_idx : int
            Index of PV row on which the above point is located
        tilted_to_left : list of bool
            Flags indicating when the PV rows are strictly tilted to the left
        is_back_side : bool
            Flag indicating if point is located on back side of PV row

        Returns
        -------
        horizon_shading_pct : np.ndarray
            Percentage vector of horizon band irradiance shading
            (from 0 to 100)
        """
        n_pvrows = len(ts_pvrows)
        if pvrow_idx == 0:
            shading_pct_left = np.zeros_like(tilted_to_left)
        else:
            high_pt_left = ts_pvrows[
                pvrow_idx - 1].full_pvrow_coords.highest_point
            shading_angle_left = np.rad2deg(np.abs(np.arctan(
                (high_pt_left.y - ts_point_coords.y)
                / (high_pt_left.x - ts_point_coords.x))))
            shading_pct_left = calculate_horizon_band_shading(
                shading_angle_left, self.horizon_band_angle)

        if pvrow_idx == (n_pvrows - 1):
            shading_pct_right = np.zeros_like(tilted_to_left)
        else:
            high_pt_right = ts_pvrows[
                pvrow_idx + 1].full_pvrow_coords.highest_point
            shading_angle_right = np.rad2deg(np.abs(np.arctan(
                (high_pt_right.y - ts_point_coords.y)
                / (high_pt_right.x - ts_point_coords.x))))
            shading_pct_right = calculate_horizon_band_shading(
                shading_angle_right, self.horizon_band_angle)

        if is_back_side:
            shading_pct = np.where(tilted_to_left, shading_pct_right,
                                   shading_pct_left)
        else:
            shading_pct = np.where(tilted_to_left, shading_pct_left,
                                   shading_pct_right)

        return shading_pct

    def _calculate_circumsolar_shading_pct(self, surface, idx_neighbor, pvrows,
                                           solar_2d_vector):
        """Model method to calculate circumsolar shading on surfaces of
        the ordered PV array.
        TODO: This needs to be merged with horizon shading for performance

        Parameters
        ----------
        surface : :py:class:`~pvfactors.geometry.base.PVSurface` object
            PV surface for which some horizon band shading will occur
        idx_neighbor : int
            Index of the neighboring PV row (can be ``None``)
        pvrows : list of :py:class:`~pvfactors.geometry.pvrow.PVRow` objects
            List of PV rows on which ``idx_neighbor`` will be used
        solar_2d_vector : list
            Solar vector in the 2D PV array representation

        Returns
        -------
        circ_shading_pct : float
            Percentage of circumsolar irradiance shaded (from 0 to 100)
        """

        # TODO: should be applied to all pvrow surfaces

        if idx_neighbor is not None:
            # Calculate the solar and circumsolar elevation angles in 2D plane
            solar_2d_elevation = np.abs(
                np.arctan(solar_2d_vector[1] / solar_2d_vector[0])
            ) * 180. / np.pi
            lower_angle_circumsolar = (solar_2d_elevation -
                                       self.circumsolar_angle / 2.)
            centroid = surface.centroid
            neighbor_point = pvrows[idx_neighbor].highest_point
            shading_angle = np.abs(np.arctan(
                (neighbor_point.y - centroid.y) /
                (neighbor_point.x - centroid.x))) * 180. / np.pi
            percentage_circ_angle_covered = (shading_angle - lower_angle_circumsolar) \
                / self.circumsolar_angle * 100.
            circ_shading_pct = calculate_circumsolar_shading(
                percentage_circ_angle_covered, model=self.circumsolar_model)

        return circ_shading_pct

    @staticmethod
    def _calculate_luminance_poa_components(
            timestamps, DNI, DHI, solar_zenith, solar_azimuth, surface_tilt,
            surface_azimuth):
        """Calculate Perez-like luminance and plane-of-array irradiance values.

        Parameters
        ----------
        timestamps : array-like
            List of timestamps of the simulation.
        DNI : array-like
            Direct normal irradiance values [W/m2]
        DHI : array-like
            Diffuse horizontal irradiance values [W/m2]
        solar_zenith : array-like
            Solar zenith angles [deg]
        solar_azimuth : array-like
            Solar azimuth angles [deg]
        surface_tilt : array-like
            Surface tilt angles, from 0 to 180 [deg]
        surface_azimuth : array-like
            Surface azimuth angles [deg]

        Returns
        -------
        luminance_isotropic : numpy array
            Luminance values of the isotropic sky dome area
        luminance_circumsolar : numpy array
            Luminance values of the circumsolar area
        poa_horizon : numpy array
            Plane-of-array irradiance coming from horizon band and incident on
            the sides of the PV rows [W/m2]
        poa_circumsolar_front : numpy array
            Plane-of-array irradiance coming from the circumsolar area and
            incident on the front side of the PV rows [W/m2]
        poa_circumsolar_back : numpy array
            Plane-of-array irradiance coming from the circumsolar area and
            incident on the back side of the PV rows [W/m2]
        aoi_front_pvrow : numpy array
            Angle of incidence of direct light on front side of PV rows [deg]
        aoi_back_pvrow : numpy array
            Angle of incidence of direct light on back side of PV rows [deg]
        """

        # Calculations from utils function
        df_inputs = perez_diffuse_luminance(
            timestamps, surface_tilt, surface_azimuth, solar_zenith,
            solar_azimuth, DNI, DHI)

        luminance_isotropic = df_inputs.luminance_isotropic.values
        luminance_circumsolar = df_inputs.luminance_circumsolar.values
        poa_horizon = df_inputs.poa_horizon.values
        poa_circumsolar_front = df_inputs.poa_circumsolar.values

        # Calculate AOI on front pvrow using pvlib implementation
        aoi_front_pvrow = aoi_function(
            surface_tilt, surface_azimuth, solar_zenith, solar_azimuth)
        aoi_back_pvrow = 180. - aoi_front_pvrow

        # Will be used for back surface adjustments: from Perez model
        # FIXME: pvlib clips the angle values to calculate vf -> adjust here
        vf_circumsolar_backsurface = cosd(aoi_back_pvrow) / cosd(solar_zenith)
        poa_circumsolar_back = (luminance_circumsolar
                                * vf_circumsolar_backsurface)

        # Return only >0 values for poa_horizon
        poa_horizon = np.abs(poa_horizon)

        return luminance_isotropic, luminance_circumsolar, poa_horizon, \
            poa_circumsolar_front, poa_circumsolar_back, \
            aoi_front_pvrow, aoi_back_pvrow

    def _calculate_faoi_modifiers(self, aoi_front_pvrow, aoi_back_pvrow,
                                  surface_tilt, albedo):
        """Calculate fAOI modifier values for all surface types: front PV row,
        back PV row, and ground.

        Parameters
        ----------
        aoi_front_pvrow : np.ndarray
            Direct light angle of incidence values for PV row front side [deg]
        aoi_back_pvrow : np.ndarray
            Direct light angle of incidence values for PV row back side [deg]
        surface_tilt : np.ndarray
            Surface tilt of PV row surfaces, measured from horizontal,
            ranging from 0 to 180 [deg]
        albedo : np.ndarray
            Ground albedo values
        """
        # Need to update aoi measurement: input measured from normal,
        # but functions require to measure it from surface horizontal
        aoi_front = np.abs(90. - aoi_front_pvrow)
        aoi_back = np.abs(90. - aoi_back_pvrow)
        # --- front
        if self.faoi_fn_front is not None:
            faoi_direct_rays = self.faoi_fn_front(aoi_front)
            self.faoi_front['direct'] = faoi_direct_rays
            # Assume that circumsolar is just a point for this
            self.faoi_front['circumsolar'] = faoi_direct_rays
            # Assume horizon band has zero width
            self.faoi_front['horizon'] = self.faoi_fn_front(surface_tilt)
        else:
            faoi_diffuse_front = ((1. - self.rho_front)
                                  * np.ones_like(surface_tilt))
            self.faoi_front['direct'] = faoi_diffuse_front
            self.faoi_front['circumsolar'] = faoi_diffuse_front
            self.faoi_front['horizon'] = faoi_diffuse_front

        # --- back
        if self.faoi_fn_back is not None:
            faoi_direct_rays = self.faoi_fn_back(aoi_back)
            self.faoi_back['direct'] = faoi_direct_rays
            # Assume that circumsolar is just a point for this
            self.faoi_back['circumsolar'] = faoi_direct_rays
            # Assume horizon band has zero width
            self.faoi_back['horizon'] = self.faoi_fn_back(surface_tilt)
        else:
            faoi_diffuse_back = ((1. - self.rho_back)
                                 * np.ones_like(surface_tilt))
            self.faoi_back['direct'] = faoi_diffuse_back
            self.faoi_back['circumsolar'] = faoi_diffuse_back
            self.faoi_back['horizon'] = faoi_diffuse_back

        # --- ground
        self.faoi_ground = (1. - albedo)



"""Classes for implementation of ground geometry"""
from pvfactors import PVFactorsError
from pvfactors.config import (
    MAX_X_GROUND, MIN_X_GROUND, Y_GROUND, DISTANCE_TOLERANCE, COLOR_DIC)
from pvfactors.geometry.base import (
    BaseSide, PVSegment, ShadeCollection, PVSurface)
from pvfactors.geometry.timeseries import (
    TsShadeCollection, TsLineCoords, TsPointCoords, TsSurface,
    _get_params_at_idx)
from shapely.geometry import LineString
import numpy as np
from copy import deepcopy


class TsGround(object):
    """Timeseries ground class: this class is a vectorized version of the
    PV ground geometry class, and it will store timeseries shaded ground
    and illuminated ground elements, as well as pv row cut points."""

    # TODO: this needs to be passed at initialization for flexibility
    x_min = MIN_X_GROUND
    x_max = MAX_X_GROUND

    def __init__(self, shadow_elements, illum_elements, param_names=None,
                 flag_overlap=None, cut_point_coords=None, y_ground=None):
        """Initialize timeseries ground using list of timeseries surfaces
        for the ground shadows

        Parameters
        ----------
        shadow_elements : \
        list of :py:class:`~pvfactors.geometry.pvground.TsGroundElement`
            Timeseries shaded ground elements
        illum_elements : \
        list of :py:class:`~pvfactors.geometry.pvground.TsGroundElement`
            Timeseries illuminated ground elements
        param_names : list of str, optional
            List of names of surface parameters to use when creating geometries
            (Default = None)
        flag_overlap : list of bool, optional
            Flags indicating if the ground shadows are overlapping, for all
            time steps (Default=None). I.e. is there direct shading on pv rows?
        cut_point_coords : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`, \
        optional
            List of cut point coordinates, as calculated for timeseries PV rows
            (Default = None)
        y_ground : float, optional
            Y coordinate of flat ground [m] (Default=None)
        """
        # Lists of timeseries ground elements
        self.shadow_elements = shadow_elements
        self.illum_elements = illum_elements
        # Shade collections
        list_shaded_surf = []
        list_illum_surf = []
        for shadow_el in shadow_elements:
            list_shaded_surf += shadow_el.all_ts_surfaces
        for illum_el in illum_elements:
            list_illum_surf += illum_el.all_ts_surfaces
        self.illum = TsShadeCollection(list_illum_surf, False)
        self.shaded = TsShadeCollection(list_shaded_surf, True)
        # Other ground attributes
        self.param_names = [] if param_names is None else param_names
        self.flag_overlap = flag_overlap
        self.cut_point_coords = [] if cut_point_coords is None \
            else cut_point_coords
        self.y_ground = y_ground
        self.shaded_params = dict.fromkeys(self.param_names)
        self.illum_params = dict.fromkeys(self.param_names)

    @classmethod
    def from_ts_pvrows_and_angles(cls, list_ts_pvrows, alpha_vec, rotation_vec,
                                  y_ground=Y_GROUND, flag_overlap=None,
                                  param_names=None):
        """Create timeseries ground from list of timeseries PV rows, and
        PV array and solar angles.

        Parameters
        ----------
        list_ts_pvrows : \
        list of :py:class:`~pvfactors.geometry.pvrow.TsPVRow`
            Timeseries PV rows to use to calculate timeseries ground shadows
        alpha_vec : np.ndarray
            Angle made by 2d solar vector and PV array x-axis [rad]
        rotation_vec : np.ndarray
            Timeseries rotation values of the PV row [deg]
        y_ground : float, optional
            Fixed y coordinate of flat ground [m] (Default = Y_GROUND constant)
        flag_overlap : list of bool, optional
            Flags indicating if the ground shadows are overlapping, for all
            time steps (Default=None). I.e. is there direct shading on pv rows?
        param_names : list of str, optional
            List of names of surface parameters to use when creating geometries
            (Default = None)
        """
        rotation_vec = np.deg2rad(rotation_vec)
        n_steps = len(rotation_vec)
        # Calculate coords of ground shadows and cutting points
        ground_shadow_coords = []
        cut_point_coords = []
        for ts_pvrow in list_ts_pvrows:
            # Get pvrow coords
            x1s_pvrow = ts_pvrow.full_pvrow_coords.b1.x
            y1s_pvrow = ts_pvrow.full_pvrow_coords.b1.y
            x2s_pvrow = ts_pvrow.full_pvrow_coords.b2.x
            y2s_pvrow = ts_pvrow.full_pvrow_coords.b2.y
            # --- Shadow coords calculation
            # Calculate x coords of shadow
            x1s_shadow = x1s_pvrow - (y1s_pvrow - y_ground) / np.tan(alpha_vec)
            x2s_shadow = x2s_pvrow - (y2s_pvrow - y_ground) / np.tan(alpha_vec)
            # Order x coords from left to right
            x1s_on_left = x1s_shadow <= x2s_shadow
            xs_left_shadow = np.where(x1s_on_left, x1s_shadow, x2s_shadow)
            xs_right_shadow = np.where(x1s_on_left, x2s_shadow, x1s_shadow)
            # Append shadow coords to list
            ground_shadow_coords.append(
                [[xs_left_shadow, y_ground * np.ones(n_steps)],
                 [xs_right_shadow, y_ground * np.ones(n_steps)]])
            # --- Cutting points coords calculation
            dx = (y1s_pvrow - y_ground) / np.tan(rotation_vec)
            cut_point_coords.append(
                TsPointCoords(x1s_pvrow - dx, y_ground * np.ones(n_steps)))

        ground_shadow_coords = np.array(ground_shadow_coords)
        return cls.from_ordered_shadows_coords(
            ground_shadow_coords, flag_overlap=flag_overlap,
            cut_point_coords=cut_point_coords, param_names=param_names,
            y_ground=y_ground)

    @classmethod
    def from_ordered_shadows_coords(cls, shadow_coords, flag_overlap=None,
                                    param_names=None, cut_point_coords=None,
                                    y_ground=Y_GROUND):
        """Create timeseries ground from list of ground shadow coordinates.

        Parameters
        ----------
        shadow_coords : np.ndarray
            List of ordered ground shadow coordinates (from left to right)
        flag_overlap : list of bool, optional
            Flags indicating if the ground shadows are overlapping, for all
            time steps (Default=None). I.e. is there direct shading on pv rows?
        param_names : list of str, optional
            List of names of surface parameters to use when creating geometries
            (Default = None)
        cut_point_coords : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`, \
        optional
            List of cut point coordinates, as calculated for timeseries PV rows
            (Default = None)
        y_ground : float, optional
            Fixed y coordinate of flat ground [m] (Default = Y_GROUND constant)
        """

        # Get cut point coords if any
        cut_point_coords = cut_point_coords or []
        # Create shadow coordinate objects
        list_shadow_coords = [TsLineCoords.from_array(coords)
                              for coords in shadow_coords]
        # If the overlap flags were passed, make sure shadows don't overlap
        if flag_overlap is not None:
            if len(list_shadow_coords) > 1:
                for idx, coords in enumerate(list_shadow_coords[:-1]):
                    coords.b2.x = np.where(flag_overlap,
                                           list_shadow_coords[idx + 1].b1.x,
                                           coords.b2.x)
        # Create shaded ground elements
        ts_shadows_elements = cls._shadow_elements_from_coords_and_cut_pts(
            list_shadow_coords, cut_point_coords, param_names)
        # Create illuminated ground elements
        ts_illum_elements = cls._illum_elements_from_coords_and_cut_pts(
            ts_shadows_elements, cut_point_coords, param_names, y_ground)
        return cls(ts_shadows_elements, ts_illum_elements,
                   param_names=param_names, flag_overlap=flag_overlap,
                   cut_point_coords=cut_point_coords, y_ground=y_ground)

    def at(self, idx, x_min_max=None, merge_if_flag_overlap=True,
           with_cut_points=True):
        """Generate a PV ground geometry for the desired index. This will
        only return non-point surfaces within the ground bounds, i.e.
        surfaces that are not points, and which are within x_min and x_max.

        Parameters
        ----------
        idx : int
            Index to use to generate PV ground geometry
        x_min_max : tuple, optional
            List of minimum and maximum x coordinates for the flat surface [m]
            (Default = None)
        merge_if_flag_overlap : bool, optional
            Decide whether to merge all shadows if they overlap or not
            (Default = True)
        with_cut_points :  bool, optional
            Decide whether to include the saved cut points in the created
            PV ground geometry (Default = True)

        Returns
        -------
        pvground : :py:class:`~pvfactors.geometry.pvground.PVGround`
        """
        # Get shadow elements that are not points at the given index
        non_pt_shadow_elements = [
            shadow_el for shadow_el in self.shadow_elements
            if shadow_el.coords.length[idx] > DISTANCE_TOLERANCE]

        if with_cut_points:
            # We want the ground surfaces broken up at the cut points
            if merge_if_flag_overlap:
                # We want to merge the shadow surfaces when they overlap
                list_shadow_surfaces = self._merge_shadow_surfaces(
                    idx, non_pt_shadow_elements)
            else:
                # No need to merge the shadow surfaces
                list_shadow_surfaces = []
                for shadow_el in non_pt_shadow_elements:
                    list_shadow_surfaces += \
                        shadow_el.non_point_surfaces_at(idx)
            # Get the illuminated surfaces
            list_illum_surfaces = []
            for illum_el in self.illum_elements:
                list_illum_surfaces += illum_el.non_point_surfaces_at(idx)
        else:
            # No need to break up the surfaces at the cut points
            # We will need to build up new surfaces (since not done by classes)

            # Get the parameters at the given index
            illum_params = _get_params_at_idx(idx, self.illum_params)
            shaded_params = _get_params_at_idx(idx, self.shaded_params)

            if merge_if_flag_overlap and (self.flag_overlap is not None):
                # We want to merge the shadow surfaces when they overlap
                is_overlap = self.flag_overlap[idx]
                if is_overlap and (len(non_pt_shadow_elements) > 1):
                    coords = [non_pt_shadow_elements[0].b1.at(idx),
                              non_pt_shadow_elements[-1].b2.at(idx)]
                    list_shadow_surfaces = [PVSurface(
                        coords, shaded=True, param_names=self.param_names,
                        params=shaded_params)]
                else:
                    # No overlap for the given index or config
                    list_shadow_surfaces = [
                        PVSurface(shadow_el.coords.at(idx),
                                  shaded=True, params=shaded_params,
                                  param_names=self.param_names)
                        for shadow_el in non_pt_shadow_elements
                        if shadow_el.coords.length[idx]
                        > DISTANCE_TOLERANCE]
            else:
                # No need to merge the shadow surfaces
                list_shadow_surfaces = [
                    PVSurface(shadow_el.coords.at(idx),
                              shaded=True, params=shaded_params,
                              param_names=self.param_names)
                    for shadow_el in non_pt_shadow_elements
                    if shadow_el.coords.length[idx]
                    > DISTANCE_TOLERANCE]
            # Get the illuminated surfaces
            list_illum_surfaces = [PVSurface(illum_el.coords.at(idx),
                                             shaded=False, params=illum_params,
                                             param_names=self.param_names)
                                   for illum_el in self.illum_elements
                                   if illum_el.coords.length[idx]
                                   > DISTANCE_TOLERANCE]

        # Pass the created lists to the PVGround builder
        return PVGround.from_lists_surfaces(
            list_shadow_surfaces, list_illum_surfaces,
            param_names=self.param_names, y_ground=self.y_ground,
            x_min_max=x_min_max)

    def plot_at_idx(self, idx, ax, color_shaded=COLOR_DIC['pvrow_shaded'],
                    color_illum=COLOR_DIC['pvrow_illum'], x_min_max=None,
                    merge_if_flag_overlap=True, with_cut_points=True,
                    with_surface_index=False):
        """Plot timeseries ground at a certain index.

        Parameters
        ----------
        idx : int
            Index to use to plot timeseries side
        ax : :py:class:`matplotlib.pyplot.axes` object
            Axes for plotting
        color_shaded : str, optional
            Color to use for plotting the shaded surfaces (Default =
            COLOR_DIC['pvrow_shaded'])
        color_shaded : str, optional
            Color to use for plotting the illuminated surfaces (Default =
            COLOR_DIC['pvrow_illum'])
        x_min_max : tuple, optional
            List of minimum and maximum x coordinates for the flat surface [m]
            (Default = None)
        merge_if_flag_overlap : bool, optional
            Decide whether to merge all shadows if they overlap or not
            (Default = True)
        with_cut_points :  bool, optional
            Decide whether to include the saved cut points in the created
            PV ground geometry (Default = True)
        with_surface_index : bool, optional
            Plot the surfaces with their index values (Default = False)
        """
        pvground = self.at(idx, x_min_max=x_min_max,
                           merge_if_flag_overlap=merge_if_flag_overlap,
                           with_cut_points=with_cut_points)
        pvground.plot(ax, color_shaded=color_shaded, color_illum=color_illum,
                      with_index=with_surface_index)

    def update_params(self, new_dict):
        """Update the illuminated parameters with new ones, not only for the
        timeseries ground, but also for its ground elements and the timeseries
        surfaces of the ground elements, so that they are all synced.

        Parameters
        ----------
        new_dict : dict
            New parameters
        """
        self.update_illum_params(new_dict)
        self.update_shaded_params(new_dict)

    def update_illum_params(self, new_dict):
        """Update the illuminated parameters with new ones, not only for the
        timeseries ground, but also for its ground elements and the timeseries
        surfaces of the ground elements, so that they are all synced.

        Parameters
        ----------
        new_dict : dict
            New parameters
        """
        self.illum_params.update(new_dict)
        for illum_el in self.illum_elements:
            illum_el.params.update(new_dict)
            for surf in illum_el.surface_list:
                surf.params.update(new_dict)

    def update_shaded_params(self, new_dict):
        """Update the shaded parameters with new ones, not only for the
        timeseries ground, but also for its ground elements and the timeseries
        surfaces of the ground elements, so that they are all synced.

        Parameters
        ----------
        new_dict : dict
            New parameters
        """
        self.shaded_params.update(new_dict)
        for shaded_el in self.shadow_elements:
            shaded_el.params.update(new_dict)
            for surf in shaded_el.surface_list:
                surf.params.update(new_dict)

    def get_param_weighted(self, param):
        """Get timeseries parameter for the ts ground, after weighting by
        surface length.

        Parameters
        ----------
        param : str
            Name of parameter

        Returns
        -------
        np.ndarray
            Weighted parameter values
        """
        return self.get_param_ww(param) / self.length

    def get_param_ww(self, param):
        """Get timeseries parameter from the ground's surfaces with weight,
        i.e. after multiplying by the surface lengths.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        np.ndarray
            Timeseries parameter values multiplied by weights

        Raises
        ------
        KeyError
            if parameter name not in a surface parameters
        """
        value = 0.
        for shadow_el in self.shadow_elements:
            value += shadow_el.get_param_ww(param)
        for illum_el in self.illum_elements:
            value += illum_el.get_param_ww(param)
        return value

    def shadow_coords_left_of_cut_point(self, idx_cut_pt):
        """Get coordinates of shadows located on the left side of the cut point
        with given index. The coordinates of the shadows will be bounded
        by the coordinates of the cut point and the default minimum
        ground x values.

        Parameters
        ----------
        idx_cut_pt : int
            Index of the cut point of interest

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Coordinates of the shadows on the left side of the cut point
        """
        cut_pt_coords = self.cut_point_coords[idx_cut_pt]
        return [shadow_el._coords_left_of_cut_point(shadow_el.coords,
                                                    cut_pt_coords)
                for shadow_el in self.shadow_elements]

    def shadow_coords_right_of_cut_point(self, idx_cut_pt):
        """Get coordinates of shadows located on the right side of the cut
        point with given index. The coordinates of the shadows will be bounded
        by the coordinates of the cut point and the default maximum
        ground x values.

        Parameters
        ----------
        idx_cut_pt : int
            Index of the cut point of interest

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Coordinates of the shadows on the right side of the cut point
        """
        cut_pt_coords = self.cut_point_coords[idx_cut_pt]
        return [shadow_el._coords_right_of_cut_point(shadow_el.coords,
                                                     cut_pt_coords)
                for shadow_el in self.shadow_elements]

    def ts_surfaces_side_of_cut_point(self, side, idx_cut_pt):
        """Get a list of all the ts ground surfaces an a request side of
        a cut point

        Parameters
        ----------
        side : str
            Side of the cut point, either 'left' or 'right'
        idx_cut_pt : int
            Index of the cut point, on whose side we want to get the ground
            surfaces

        Returns
        -------
        list
            List of timeseries ground surfaces on the side of the cut point
        """
        list_ts_surfaces = []
        for shadow_el in self.shadow_elements:
            list_ts_surfaces += shadow_el.surface_dict[idx_cut_pt][side]
        for illum_el in self.illum_elements:
            list_ts_surfaces += illum_el.surface_dict[idx_cut_pt][side]
        return list_ts_surfaces

    @property
    def n_ts_surfaces(self):
        """Number of timeseries surfaces in the ts ground"""
        return self.n_ts_shaded_surfaces + self.n_ts_illum_surfaces

    @property
    def n_ts_shaded_surfaces(self):
        """Number of shaded timeseries surfaces in the ts ground"""
        n_ts_surfaces = 0
        for shadow_el in self.shadow_elements:
            n_ts_surfaces += shadow_el.n_ts_surfaces
        return n_ts_surfaces

    @property
    def n_ts_illum_surfaces(self):
        """Number of illuminated timeseries surfaces in the ts ground"""
        n_ts_surfaces = 0
        for illum_el in self.illum_elements:
            n_ts_surfaces += illum_el.n_ts_surfaces
        return n_ts_surfaces

    @property
    def all_ts_surfaces(self):
        """Number of timeseries surfaces in the ts ground"""
        all_ts_surfaces = []
        for shadow_el in self.shadow_elements:
            all_ts_surfaces += shadow_el.all_ts_surfaces
        for illum_el in self.illum_elements:
            all_ts_surfaces += illum_el.all_ts_surfaces
        return all_ts_surfaces

    @property
    def length(self):
        """Length of the timeseries ground"""
        length = 0
        for shadow_el in self.shadow_elements:
            length += shadow_el.length
        for illum_el in self.illum_elements:
            length += illum_el.length
        return length

    @property
    def shaded_length(self):
        """Length of the timeseries ground"""
        length = 0
        for shadow_el in self.shadow_elements:
            length += shadow_el.length
        return length

    def non_point_shaded_surfaces_at(self, idx):
        """Return a list of shaded surfaces, that are not points
        at given index

        Parameters
        ----------
        idx : int
            Index at which we want the surfaces not to be points

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
        """
        list_surfaces = []
        for shadow_el in self.shadow_elements:
            list_surfaces += shadow_el.non_point_surfaces_at(0)
        return list_surfaces

    def non_point_illum_surfaces_at(self, idx):
        """Return a list of illuminated surfaces, that are not
        points at given index

        Parameters
        ----------
        idx : int
            Index at which we want the surfaces not to be points

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
        """
        list_surfaces = []
        for illum_el in self.illum_elements:
            list_surfaces += illum_el.non_point_surfaces_at(0)
        return list_surfaces

    def non_point_surfaces_at(self, idx):
        """Return a list of all surfaces that are not
        points at given index

        Parameters
        ----------
        idx : int
            Index at which we want the surfaces not to be points

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
        """
        return self.non_point_illum_surfaces_at(idx) \
            + self.non_point_shaded_surfaces_at(idx)

    def n_non_point_surfaces_at(self, idx):
        """Return the number of :py:class:`~pvfactors.geometry.base.PVSurface`
        that are not points at given index

        Parameters
        ----------
        idx : int
            Index at which we want the surfaces not to be points

        Returns
        -------
        int
        """
        return len(self.non_point_surfaces_at(idx))

    @staticmethod
    def _shadow_elements_from_coords_and_cut_pts(
            list_shadow_coords, cut_point_coords, param_names):
        """Create ground shadow elements from a list of ordered shadow
        coordinates (from left to right), and the ground cut point coordinates.

        Notes
        -----
        This method will clip the shadow coords to the limit of ground,
        i.e. the shadow coordinates shouldn't be outside of the range
        [MIN_X_GROUND, MAX_X_GROUND].

        Parameters
        ----------
        list_shadow_coords : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            List of ordered ground shadow coordinates (from left to right)
        cut_point_coords : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            List of cut point coordinates (from left to right)
        param_names : list
            List of parameter names for the ground elements

        Returns
        -------
        list_shadow_elements : \
        list of :py:class:`~pvfactors.geometry.pvground.TsGroundElement`
            Ordered list of shadow elements (from left to right)
        """

        list_shadow_elements = []
        # FIXME: x_min and x_max should be passed as inputs
        for shadow_coords in list_shadow_coords:
            shadow_coords.b1.x = np.clip(shadow_coords.b1.x, MIN_X_GROUND,
                                         MAX_X_GROUND)
            shadow_coords.b2.x = np.clip(shadow_coords.b2.x, MIN_X_GROUND,
                                         MAX_X_GROUND)
            list_shadow_elements.append(
                TsGroundElement(shadow_coords,
                                list_ordered_cut_pts_coords=cut_point_coords,
                                param_names=param_names, shaded=True))

        return list_shadow_elements

    @staticmethod
    def _illum_elements_from_coords_and_cut_pts(
            list_shadow_elements, cut_pt_coords, param_names, y_ground):
        """Create ground illuminated elements from a list of ordered shadow
        elements (from left to right), and the ground cut point coordinates.
        This method will make sure that the illuminated ground elements are
        all within the ground limits [MIN_X_GROUND, MAX_X_GROUND].

        Parameters
        ----------
        list_shadow_coords : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            List of ordered ground shadow coordinates (from left to right)
        cut_point_coords : \
        list of :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            List of cut point coordinates (from left to right)
        param_names : list
            List of parameter names for the ground elements

        Returns
        -------
        list_shadow_elements : \
        list of :py:class:`~pvfactors.geometry.pvground.TsGroundElement`
            Ordered list of shadow elements (from left to right)
        """

        list_illum_elements = []
        if len(list_shadow_elements) == 0:
            msg = """There must be at least one shadow element on the ground,
            otherwise it probably means that no PV rows were created, so
            there's no point in running a simulation..."""
            raise PVFactorsError(msg)
        n_steps = len(list_shadow_elements[0].coords.b1.x)
        y_ground_vec = y_ground * np.ones(n_steps)
        # FIXME: x_min and x_max should be passed as inputs
        next_x = MIN_X_GROUND * np.ones(n_steps)
        # Build the groud elements from left to right, starting at x_min
        # and covering the ground with illuminated elements where there's no
        # shadow
        for shadow_element in list_shadow_elements:
            x1 = next_x
            x2 = shadow_element.coords.b1.x
            coords = TsLineCoords.from_array(
                np.array([[x1, y_ground_vec], [x2, y_ground_vec]]))
            list_illum_elements.append(TsGroundElement(
                coords, list_ordered_cut_pts_coords=cut_pt_coords,
                param_names=param_names, shaded=False))
            next_x = shadow_element.coords.b2.x
        # Add the last illuminated element to the list
        coords = TsLineCoords.from_array(
            np.array([[next_x, y_ground_vec],
                      [MAX_X_GROUND * np.ones(n_steps), y_ground_vec]]))
        list_illum_elements.append(TsGroundElement(
            coords, list_ordered_cut_pts_coords=cut_pt_coords,
            param_names=param_names, shaded=False))

        return list_illum_elements

    def _merge_shadow_surfaces(self, idx, non_pt_shadow_elements):
        """Merge the shadow surfaces in a list of shadow elements
        at the shadow boundaries only, at a given index, but keep the shadow
        surfaces broken up at the cut points.

        Parameters
        ----------
        idx : int
            Index at which we want to merge the surfaces
        non_pt_shadow_elements : \
        list of :py:class:`~pvfactors.geometry.pvground.TsGroundElement`
            List of non point shadow elements

        Returns
        -------
        list_shadow_surfaces : \
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
            List of shadow surfaces at a given index
            (ordered from left to right)
        """
        # TODO: check if it would be faster to merge the ground elements first,
        # and then break it down with the cut points

        # Decide whether to merge all shadows or not
        list_shadow_surfaces = []
        if self.flag_overlap is not None:
            # Get the overlap flags
            is_overlap = self.flag_overlap[idx]
            n_shadow_elements = len(non_pt_shadow_elements)
            if is_overlap and (n_shadow_elements > 1):
                # If there's only one shadow, not point in going through this

                # Now go from left to right and merge shadow surfaces
                surface_to_merge = None
                for i_el, shadow_el in enumerate(non_pt_shadow_elements):
                    surfaces = shadow_el.non_point_surfaces_at(idx)
                    n_surf = len(surfaces)
                    for i_surf, surface in enumerate(surfaces):
                        if i_surf == n_surf - 1:
                            # last surface, could also be first
                            if i_surf == 0:
                                # Need to merge with preceding if exists
                                if surface_to_merge is not None:
                                    coords = [surface_to_merge.boundary[0],
                                              surface.boundary[1]]
                                    surface = PVSurface(
                                        coords, shaded=True,
                                        param_names=self.param_names,
                                        params=surface.params,
                                        index=surface.index)
                            if i_el == n_shadow_elements - 1:
                                # last surface of last shadow element
                                list_shadow_surfaces.append(surface)
                            else:
                                # keep for merging with next element
                                surface_to_merge = surface
                        elif i_surf == 0:
                            # first surface but definitely not last either
                            if surface_to_merge is not None:
                                coords = [surface_to_merge.boundary[0],
                                          surface.boundary[1]]
                                list_shadow_surfaces.append(
                                    PVSurface(coords, shaded=True,
                                              param_names=self.param_names,
                                              params=surface.params,
                                              index=surface.index))
                            else:
                                list_shadow_surfaces.append(surface)
                        else:
                            # not first nor last surface
                            list_shadow_surfaces.append(surface)
            else:
                # There's no need to merge anything
                for shadow_el in non_pt_shadow_elements:
                    list_shadow_surfaces += \
                        shadow_el.non_point_surfaces_at(idx)
        else:
            # There's no need to merge anything
            for shadow_el in non_pt_shadow_elements:
                list_shadow_surfaces += shadow_el.non_point_surfaces_at(idx)

        return list_shadow_surfaces


class TsGroundElement(object):
    """Special class for timeseries ground elements: a ground element has known
    timeseries coordinate boundaries, but it will also have a break down of
    its area into n+1 timeseries surfaces located in the n+1 ground zones
    defined by the n ground cutting points.
    This is crucial to calculate view factors in a vectorized way."""

    def __init__(self, coords, list_ordered_cut_pts_coords=None,
                 param_names=None, shaded=False):
        """Initialize the timeseries ground element using its timeseries
        line coordinates, and build the timeseries surfaces for all the
        cut point zones.

        Parameters
        ----------
        coords : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates of the ground element
        list_ordered_cut_pts_coords : list, optional
            List of all the cut point timeseries coordinates
            (Default = [])
        param_names : list of str, optional
            List of names of surface parameters to use when creating geometries
            (Default = None)
        shaded : bool, optional
            Flag specifying is element is a shadow or not (Default = False)
        """
        self.coords = coords
        self.param_names = param_names or []
        self.params = dict.fromkeys(self.param_names)
        self.shaded = shaded
        self.surface_dict = None  # will be necessary for view factor calcs
        self.surface_list = []  # will be necessary for vf matrix formation
        list_ordered_cut_pts_coords = list_ordered_cut_pts_coords or []
        if len(list_ordered_cut_pts_coords) > 0:
            self._create_all_ts_surfaces(list_ordered_cut_pts_coords)
        self.n_ts_surfaces = len(self.surface_list)

    @property
    def b1(self):
        """Timeseries coordinates of first boundary point"""
        return self.coords.b1

    @property
    def b2(self):
        """Timeseries coordinates of second boundary point"""
        return self.coords.b2

    @property
    def centroid(self):
        """Timeseries point coordinates of the element's centroid"""
        return self.coords.centroid

    @property
    def length(self):
        """Timeseries length of the ground"""
        return self.coords.length

    @property
    def all_ts_surfaces(self):
        """List of all ts surfaces making up the ts ground element"""
        return self.surface_list

    def surfaces_at(self, idx):
        """Return list of surfaces (from left to right) at given index that
        make up the ground element.

        Parameters
        ----------
        idx : int
            Index of interest

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
        """
        return [surface.at(idx)
                for surface in self.surface_list]

    def non_point_surfaces_at(self, idx):
        """Return list of non-point surfaces (from left to right) at given
        index that make up the ground element.

        Parameters
        ----------
        idx : int
            Index of interest

        Returns
        -------
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
        """
        return [surface.at(idx)
                for surface in self.surface_list
                if surface.length[idx] > DISTANCE_TOLERANCE]

    def get_param_weighted(self, param):
        """Get timeseries parameter for the ground element, after weighting by
        surface length.

        Parameters
        ----------
        param : str
            Name of parameter

        Returns
        -------
        np.ndarray
            Weighted parameter values
        """
        return self.get_param_ww(param) / self.length

    def get_param_ww(self, param):
        """Get timeseries parameter from the ground element with weight,
        i.e. after multiplying by the surface lengths.

        Parameters
        ----------
        param: str
            Surface parameter to return

        Returns
        -------
        np.ndarray
            Timeseries parameter values multiplied by weights

        Raises
        ------
        KeyError
            if parameter name not in a surface parameters
        """
        value = 0.
        for ts_surf in self.surface_list:
            value += ts_surf.length * ts_surf.get_param(param)
        return value

    def _create_all_ts_surfaces(self, list_ordered_cut_pts):
        """Create all the n+1 timeseries surfaces that make up the timeseries
        ground element, and which are located in the n+1 zones defined by
        the n cut points.

        Parameters
        ----------
        list_ordered_cut_pts : list of :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            List of timeseries coordinates of all cut points, ordered from
            left to right
        """
        # Initialize dict
        self.surface_dict = {i: {'right': [], 'left': []}
                             for i in range(len(list_ordered_cut_pts))}
        n_cut_pts = len(list_ordered_cut_pts)

        next_coords = self.coords
        for idx_pt, cut_pt_coords in enumerate(list_ordered_cut_pts):
            # Get coords on left of cut pt
            coords_left = self._coords_left_of_cut_point(next_coords,
                                                         cut_pt_coords)
            # Save that surface in the required structures
            surface_left = TsSurface(coords_left, param_names=self.param_names,
                                     shaded=self.shaded)
            self.surface_list.append(surface_left)
            for i in range(idx_pt, n_cut_pts):
                self.surface_dict[i]['left'].append(surface_left)
            for j in range(0, idx_pt):
                self.surface_dict[j]['right'].append(surface_left)
            next_coords = self._coords_right_of_cut_point(next_coords,
                                                          cut_pt_coords)
        # Save the right most portion
        next_surface = TsSurface(next_coords, param_names=self.param_names,
                                 shaded=self.shaded)
        self.surface_list.append(next_surface)
        for j in range(0, n_cut_pts):
            self.surface_dict[j]['right'].append(next_surface)

    @staticmethod
    def _coords_right_of_cut_point(coords, cut_pt_coords):
        """Calculate timeseries line coordinates that are right of the given
        cut point coordinates, but still within the ground area

        Parameters
        ----------
        coords : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Original timeseries coordinates
        cut_pt_coords :
        :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of cut point
        Returns
        -------
        :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates that are located right of the cut
            point
        """
        coords = deepcopy(coords)
        # FIXME: should be using x_min x_max inputs instead of global constant
        coords.b1.x = np.maximum(coords.b1.x, cut_pt_coords.x)
        coords.b1.x = np.minimum(coords.b1.x, MAX_X_GROUND)
        coords.b2.x = np.maximum(coords.b2.x, cut_pt_coords.x)
        coords.b2.x = np.minimum(coords.b2.x, MAX_X_GROUND)
        return coords

    @staticmethod
    def _coords_left_of_cut_point(coords, cut_pt_coords):
        """Calculate timeseries line coordinates that are left of the given
        cut point coordinates, but still within the ground area

        Parameters
        ----------
        coords : :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Original timeseries coordinates
        cut_pt_coords :
        :py:class:`~pvfactors.geometry.timeseries.TsPointCoords`
            Timeseries coordinates of cut point
        Returns
        -------
        :py:class:`~pvfactors.geometry.timeseries.TsLineCoords`
            Timeseries line coordinates that are located left of the cut
            point
        """
        coords = deepcopy(coords)
        # FIXME: should be using x_min x_max inputs instead of global constant
        coords.b1.x = np.minimum(coords.b1.x, cut_pt_coords.x)
        coords.b1.x = np.maximum(coords.b1.x, MIN_X_GROUND)
        coords.b2.x = np.minimum(coords.b2.x, cut_pt_coords.x)
        coords.b2.x = np.maximum(coords.b2.x, MIN_X_GROUND)
        return coords


class PVGround(BaseSide):
    """Class that defines the ground geometry in PV arrays."""

    def __init__(self, list_segments=None, original_linestring=None):
        """Initialize PV ground geometry.

        Parameters
        ----------
        list_segments : list of :py:class:`~pvfactors.geometry.base.PVSegment`, optional
            List of PV segments that will constitute the ground (Default = [])
        original_linestring : :py:class:`shapely.geometry.LineString`, optional
            Full continuous linestring that the ground will be made of
            (Default = None)
        """
        list_segments = list_segments or []
        self.original_linestring = original_linestring
        super(PVGround, self).__init__(list_segments)

    @classmethod
    def as_flat(cls, x_min_max=None, shaded=False, y_ground=Y_GROUND,
                param_names=None):
        """Build a horizontal flat ground surface, made of 1 PV segment.

        Parameters
        ----------
        x_min_max : tuple, optional
            List of minimum and maximum x coordinates for the flat surface [m]
            (Default = None)
        shaded : bool, optional
            Shaded status of the created PV surfaces (Default = False)
        y_ground : float, optional
            Location of flat ground on y axis in [m] (Default = Y_GROUND)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = [])

        Returns
        -------
        PVGround object
        """
        param_names = param_names or []
        # Get ground boundaries
        if x_min_max is None:
            x_min, x_max = MIN_X_GROUND, MAX_X_GROUND
        else:
            x_min, x_max = x_min_max
        # Create PV segment for flat ground
        coords = [(x_min, y_ground), (x_max, y_ground)]
        seg = PVSegment.from_linestring_coords(coords, shaded=shaded,
                                               normal_vector=[0., 1.],
                                               param_names=param_names)
        return cls(list_segments=[seg], original_linestring=LineString(coords))

    @classmethod
    def from_lists_surfaces(
            cls, list_shaded_surfaces, list_illum_surfaces, x_min_max=None,
            y_ground=Y_GROUND, param_names=None):
        """Create ground from lists of shaded and illuminated PV surfaces.

        Parameters
        ----------
        list_shaded_surfaces : \
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
            List of shaded ground PV surfaces
        list_illum_surfaces : \
        list of :py:class:`~pvfactors.geometry.base.PVSurface`
            List of illuminated ground PV surfaces
        x_min_max : tuple, optional
            List of minimum and maximum x coordinates for the flat surface [m]
            (Default = None)
        y_ground : float, optional
            Location of flat ground on y axis in [m] (Default = Y_GROUND)
        param_names : list of str, optional
            Names of the surface parameters, eg reflectivity, total incident
            irradiance, temperature, etc. (Default = [])

        Returns
        -------
        PVGround object
        """
        param_names = param_names or []
        # Get ground boundaries
        if x_min_max is None:
            x_min, x_max = MIN_X_GROUND, MAX_X_GROUND
        else:
            x_min, x_max = x_min_max
        full_extent_coords = [(x_min, y_ground), (x_max, y_ground)]

        # Create the shade collections
        shaded_collection = ShadeCollection(
            list_surfaces=list_shaded_surfaces, shaded=True,
            param_names=param_names)
        illum_collection = ShadeCollection(
            list_surfaces=list_illum_surfaces, shaded=False,
            param_names=param_names)

        # Create the ground segment
        segment = PVSegment(illum_collection=illum_collection,
                            shaded_collection=shaded_collection)

        return cls(list_segments=[segment],
                   original_linestring=LineString(full_extent_coords))

    @property
    def boundary(self):
        """Boundaries of the ground's original linestring."""
        return self.original_linestring.boundary



"""Module containing examples of report builder functions and classes."""


def example_fn_build_report(pvarray):
    """Example function that builds a report when used in the
    :py:class:`~pvfactors.engine.PVEngine` with full or fast mode simulations.
    Here it will be a dictionary with lists of calculated values.

    Parameters
    ----------
    pvarray : PV array object
        PV array with updated calculation values

    Returns
    -------
    report : dict
        Report updated with newly calculated values
    """
    return {'qinc_front': pvarray.ts_pvrows[1].front
            .get_param_weighted('qinc').tolist(),
            'qinc_back': pvarray.ts_pvrows[1].back
            .get_param_weighted('qinc').tolist(),
            'iso_front': pvarray.ts_pvrows[1].front
            .get_param_weighted('isotropic').tolist(),
            'iso_back': pvarray.ts_pvrows[1].back
            .get_param_weighted('isotropic').tolist()}


class ExampleReportBuilder(object):
    """A class is required to build reports when running calculations with
    multiprocessing because of python constraints"""

    @staticmethod
    def build(pvarray):
        """Method that will build the simulation report. Here we're using the
        previously defined
        :py:function:`~pvfactors.report.example_fn_build_report`.

        Parameters
        ----------
        pvarray : PV array object
            PV array with updated calculation values

        Returns
        -------
        report : dict
            Report updated with newly calculated values
        """
        return example_fn_build_report(pvarray)

    @staticmethod
    def merge(reports):
        """Method used to merge multiple reports together. Here it simply
        concatenates the lists of values saved in the different reports.

        Parameters
        ----------
        reports : list of dict
            List of reports that need to be concatenated together

        Returns
        -------
        report : dict
            Final report with all concatenated values
        """
        report = reports[0]
        # merge other reports if any
        keys_report = list(reports[0].keys())
        for other_report in reports[1:]:
            for key in keys_report:
                report[key] += other_report[key]
        return report



"""Base functions used to plot 2D PV geometries"""


def plot_coords(ax, ob):
    """Plot coordinates of shapely objects

    Parameters
    ----------
    ax : ``matplotlib.pyplot.Axes`` object
        Axes for plotting
    ob : ``Shapely`` object
        Geometry object whose x,y coordinates should be plotted

    """
    try:
        x, y = ob.xy
        ax.plot(x, y, 'o', color='#999999', zorder=1)
    except NotImplementedError:
        for line in ob:
            x, y = line.xy
            ax.plot(x, y, 'o', color='#999999', zorder=1)


def plot_bounds(ax, ob):
    """Plot boundaries of shapely object

    Parameters
    ----------
    ax : ``matplotlib.pyplot.Axes`` object
        Axes for plotting
    ob : ``Shapely`` object
        Geometry object whose boundaries should be plotted

    """
    # Check if shadow reduces to one point (for very specific sun alignment)
    if len(ob.boundary) == 0:
        x, y = ob.coords[0]
    else:
        x, y = zip(*list((p.x, p.y) for p in ob.boundary))
    ax.plot(x, y, 'o', color='#000000', zorder=1)


def plot_line(ax, ob, line_color):
    """Plot boundaries of shapely line

    Parameters
    ----------
    ax : ``matplotlib.pyplot.Axes`` object
        Axes for plotting
    ob : ``Shapely`` object
        Geometry object whose boundaries should be plotted
    line_color : str
        matplotlib color to use for plotting the line

    """
    try:
        x, y = ob.xy
        ax.plot(x, y, color=line_color, alpha=0.7,
                linewidth=3, solid_capstyle='round', zorder=2)
    except NotImplementedError:
        for line in ob:
            x, y = line.xy
            ax.plot(x, y, color=line_color,
                    alpha=0.7, linewidth=3, solid_capstyle='round', zorder=2)
